# Tai Sakuma <tai.sakuma@gmail.com>
import inspect
import pytest

import alphatwirl

##__________________________________________________________________||
has_no_pandas = False
try:
    import pandas
except ImportError:
    has_no_pandas = True

##__________________________________________________________________||
has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

##__________________________________________________________________||
@pytest.mark.skipif(has_no_pandas, reason="has no pandas")
def test_with_pandas():
    assert inspect.isclass(alphatwirl.collector.ToDataFrameWithDatasetColumn)
    assert inspect.isclass(alphatwirl.collector.WritePandasDataFrameToFile)

@pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")
def test_with_ROOT():
    assert inspect.isclass(alphatwirl.roottree.BEvents)
    assert inspect.isclass(alphatwirl.roottree.BranchAddressManager)
    assert inspect.isclass(alphatwirl.roottree.BranchAddressManagerForVector)

def test_functions():
    assert inspect.isfunction(alphatwirl.mkdir_p)
    assert inspect.isfunction(alphatwirl.configure.build_counter_collector_pair)
    assert inspect.isfunction(alphatwirl.configure.build_progressMonitor_communicationChannel)

def test_classes():
    assert inspect.isclass(alphatwirl.binning.Binning)
    assert inspect.isclass(alphatwirl.binning.Echo)
    assert inspect.isclass(alphatwirl.binning.Round)
    assert inspect.isclass(alphatwirl.binning.RoundLog)
    assert inspect.isclass(alphatwirl.binning.Combine)
    assert inspect.isclass(alphatwirl.collector.ToTupleListWithDatasetColumn)
    assert inspect.isclass(alphatwirl.collector.WriteListToFile)
    assert inspect.isclass(alphatwirl.concurrently.CommunicationChannel)
    assert inspect.isclass(alphatwirl.concurrently.CommunicationChannel0)
    assert inspect.isclass(alphatwirl.concurrently.TaskPackage)
    assert inspect.isclass(alphatwirl.concurrently.TaskPackageDropbox)
    assert inspect.isclass(alphatwirl.concurrently.MultiprocessingDropbox)
    assert inspect.isclass(alphatwirl.concurrently.SubprocessRunner)
    assert inspect.isclass(alphatwirl.concurrently.WorkingArea)
    assert inspect.isclass(alphatwirl.concurrently.HTCondorJobSubmitter)
    assert inspect.isclass(alphatwirl.configure.TableConfigCompleter)
    assert inspect.isclass(alphatwirl.configure.TableFileNameComposer)
    assert inspect.isclass(alphatwirl.roottree.Branch)
    assert inspect.isclass(alphatwirl.roottree.Events)
    assert inspect.isclass(alphatwirl.summary.Reader)
    assert inspect.isclass(alphatwirl.summary.Count)
    assert inspect.isclass(alphatwirl.summary.NextKeyComposer)
    assert inspect.isclass(alphatwirl.loop.Collector)
    assert inspect.isclass(alphatwirl.loop.CollectorComposite)
    assert inspect.isclass(alphatwirl.loop.DatasetIntoEventBuildersSplitter)
    assert inspect.isclass(alphatwirl.loop.EventLoop)
    assert inspect.isclass(alphatwirl.loop.EventLoopProgressReportWriter)
    assert inspect.isclass(alphatwirl.loop.EventLoopRunner)
    assert inspect.isclass(alphatwirl.loop.EventsInDatasetReader)
    assert inspect.isclass(alphatwirl.loop.MPEventLoopRunner)
    assert inspect.isclass(alphatwirl.loop.NullCollector)
    assert inspect.isclass(alphatwirl.loop.ReaderComposite)
    assert inspect.isclass(alphatwirl.progressbar.BProgressMonitor)
    assert inspect.isclass(alphatwirl.progressbar.NullProgressMonitor)
    assert inspect.isclass(alphatwirl.progressbar.ProgressBar)
    assert inspect.isclass(alphatwirl.progressbar.ProgressMonitor)
    assert inspect.isclass(alphatwirl.progressbar.ProgressMonitor)
    assert inspect.isclass(alphatwirl.progressbar.ProgressPrint)
    assert inspect.isclass(alphatwirl.progressbar.ProgressReport)
    assert inspect.isclass(alphatwirl.progressbar.ProgressReportPickup)
    assert inspect.isclass(alphatwirl.progressbar.ProgressReporter)
    assert inspect.isclass(alphatwirl.summary.BackrefMultipleArrayReader)
    assert inspect.isclass(alphatwirl.summary.Count)
    assert inspect.isclass(alphatwirl.summary.KeyValueComposer)
    assert inspect.isclass(alphatwirl.summary.NextKeyComposer)
    assert inspect.isclass(alphatwirl.summary.Reader)
    assert inspect.isclass(alphatwirl.summary.Scan)
    assert inspect.isclass(alphatwirl.summary.Sum)
    assert inspect.isclass(alphatwirl.summary.Summarizer)
    assert inspect.isclass(alphatwirl.summary.WeightCalculatorOne)

##________________________________._________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import numbers
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import inspect_tree

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def tree():
    input_file_name = 'sample_03.root'
    tree_name = 'tree'
    input_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', input_file_name)

    input_file = ROOT.TFile.Open(input_path)
    yield input_file.Get(tree_name)


def test_inspect(tree):
    actual = inspect_tree(tree)

    ## assert expected == actual ## this doesn't work in different
                                 ## architectures.

    assert len(expected) == len(actual) # only 'leaves'
    leaves_expected = expected['leaves']
    leaves_actual = actual['leaves']
    assert len(leaves_expected) == len(leaves_actual)
    for leave_expected, leave_actual in zip(leaves_expected, leaves_actual):
        assert len(leave_expected) == len(leave_actual)
        keys = list(leave_expected.keys())
        for key in keys:
            e = leave_expected[key]
            a = leave_actual[key]
            if isinstance(e, numbers.Integral):
                assert e == a
            elif isinstance(e, numbers.Number):
                assert e == pytest.approx(a, rel=1e-1)
            else:
                assert e == a

expected = {
    'leaves': [
        {'compression_factor': 7.472972972972973,
         'countname': None,
         'isarray': '0',
         'name': 'bChar',
         'size': 7.05718994140625e-05,
         'title': 'bChar/B',
         'type': 'Char_t',
         'uncompressed_size': 0.0005273818969726562},
        {'compression_factor': 7.44,
         'countname': None,
         'isarray': '0',
         'name': 'bUChar',
         'size': 7.152557373046875e-05,
         'title': 'bUChar/b',
         'type': 'UChar_t',
         'uncompressed_size': 0.0005321502685546875},
        {'compression_factor': 7.298701298701299,
         'countname': None,
         'isarray': '0',
         'name': 'bShort',
         'size': 7.343292236328125e-05,
         'title': 'bShort/S',
         'type': 'Short_t',
         'uncompressed_size': 0.0005359649658203125},
        {'compression_factor': 7.269230769230769,
         'countname': None,
         'isarray': '0',
         'name': 'bUShort',
         'size': 7.43865966796875e-05,
         'title': 'bUShort/s',
         'type': 'UShort_t',
         'uncompressed_size': 0.0005407333374023438},
        {'compression_factor': 7.0886075949367084,
         'countname': None,
         'isarray': '0',
         'name': 'bInt',
         'size': 7.534027099609375e-05,
         'title': 'bInt/I',
         'type': 'Int_t',
         'uncompressed_size': 0.0005340576171875},
        {'compression_factor': 7.0625,
         'countname': None,
         'isarray': '0',
         'name': 'bUInt',
         'size': 7.62939453125e-05,
         'title': 'bUInt/i',
         'type': 'UInt_t',
         'uncompressed_size': 0.0005388259887695312},
        {'compression_factor': 7.037037037037037,
         'countname': None,
         'isarray': '0',
         'name': 'bFloat',
         'size': 7.724761962890625e-05,
         'title': 'bFloat/F',
         'type': 'Float_t',
         'uncompressed_size': 0.0005435943603515625},
        {'compression_factor': 6.566666666666666,
         'countname': None,
         'isarray': '0',
         'name': 'bDouble',
         'size': 8.58306884765625e-05,
         'title': 'bDouble/D',
         'type': 'Double_t',
         'uncompressed_size': 0.0005636215209960938},
        {'compression_factor': 6.566666666666666,
         'countname': None,
         'isarray': '0',
         'name': 'bLong64',
         'size': 8.58306884765625e-05,
         'title': 'bLong64/L',
         'type': 'Long64_t',
         'uncompressed_size': 0.0005636215209960938},
        {'compression_factor': 6.549450549450549,
         'countname': None,
         'isarray': '0',
         'name': 'bULong64',
         'size': 8.678436279296875e-05,
         'title': 'bULong64/l',
         'type': 'ULong64_t',
         'uncompressed_size': 0.000568389892578125},
        {'compression_factor': 7.472972972972973,
         'countname': None,
         'isarray': '0',
         'name': 'bBool',
         'size': 7.05718994140625e-05,
         'title': 'bBool/O',
         'type': 'Bool_t',
         'uncompressed_size': 0.0005273818969726562}]}

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from alphatwirl.configure import TableFileNameComposer

##__________________________________________________________________||
@pytest.mark.parametrize('init_arg, call_arg, expected', [
    pytest.param(
        { }, dict(columnNames=('var1', 'var2', 'var3')),
        'tbl_n.var1.var2.var3.txt',
        id='no-indices'),
    pytest.param(
        { },
        dict(
            columnNames=('var1', 'var2', 'var3'),
            indices=(1, None, 2)
        ),
        'tbl_n.var1-1.var2.var3-2.txt',
        id='simple'),
    pytest.param(
        dict(default_prefix='tbl_Sum'),
        dict(
            columnNames=('var1', 'var2', 'var3'),
            indices=(1, None, 2)
        ),
        'tbl_Sum.var1-1.var2.var3-2.txt',
        id='default-prefix'),
    pytest.param(
        dict(default_suffix='hdf5'),
        dict(
            columnNames=('var1', 'var2', 'var3'),
            indices=(1, None, 2)
        ),
        'tbl_n.var1-1.var2.var3-2.hdf5',
        id='default-suffix'),
    pytest.param(
        dict(), dict(columnNames=( ), indices=( )), 'tbl_n.txt',
        id='empty'),
    pytest.param(
        dict(), dict(
            columnNames=('var1', 'var2', 'var3'),
            indices=(1, None, '*')
        ),
        'tbl_n.var1-1.var2.var3-w.txt',
        id='star'),
    pytest.param(
        dict(), dict(
            columnNames=('var1', 'var2', 'var3', 'var4', 'var5'),
            indices=(1, None, '*', '(*)', '\\1')
        ),
        'tbl_n.var1-1.var2.var3-w.var4-wp.var5-b1.txt',
        id='backref'),
    pytest.param(
        dict(default_var_separator='#'), dict(
            columnNames=('var1', 'var2', 'var3', 'var4', 'var5'),
            indices=(1, None, '*', '(*)', '\\1')
        ),
        'tbl_n#var1-1#var2#var3-w#var4-wp#var5-b1.txt',
        id='default-var-separator'),
    pytest.param(
        dict(default_idx_separator='#'), dict(
            columnNames=('var1', 'var2', 'var3', 'var4', 'var5'),
            indices=(1, None, '*', '(*)', '\\1')
        ),
        'tbl_n.var1#1.var2.var3#w.var4#wp.var5#b1.txt',
        id='default-idx-separator'),
])
def test_complete(init_arg, call_arg, expected):
    obj = TableFileNameComposer(**init_arg)
    actual = obj(**call_arg)
    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import functools

##__________________________________________________________________||
class TableFileNameComposer(object):
    """Compose a name of a file to store the table from the column names
       and indices.

    For example, if column names are 'var1', 'var2', and 'var3' and
    indices are 1, None and 2, the file name will be
    'tbl_n_component.var1-1.var2.var3-2.txt'

    """
    def __init__(self,
                 default_prefix='tbl_n',
                 default_suffix='txt',
                 default_var_separator='.',
                 default_idx_separator='-'):

        self._imp = functools.partial(
            _imp, prefix=default_prefix, suffix=default_suffix,
            var_separator=default_var_separator,
            idx_separator=default_idx_separator)

    def __call__(self, columnNames, indices=None,
                 prefix=None, suffix=None,
                 var_separator=None, idx_separator=None):

        kwargs = { }
        if prefix is not None:
            kwargs['prefix'] = prefix
        if suffix is not suffix:
            kwargs['suffix'] = suffix
        if var_separator is not None:
            kwargs['var_separator'] = var_separator
        if idx_separator is not None:
            kwargs['idx_separator'] = idx_separator

        return self._imp(columnNames, indices=indices, **kwargs)


def _imp(columnNames, indices=None, prefix=None, suffix=None,
         var_separator=None, idx_separator=None):

    if not columnNames:
        return prefix + '.' + suffix # e.g. "tbl_n_component.txt"

    if indices is None:
        colidxs = columnNames
        # e.g., ('var1', 'var2', 'var3'),

        middle = var_separator.join(colidxs)
        # e.g., 'var1.var2.var3'

        ret = prefix + var_separator + middle + '.' + suffix
        # e.g., 'tbl_n_component.var1.var2.var3.txt'

        return ret

    # e.g.,
    # columnNames = ('var1', 'var2', 'var3', 'var4', 'var5'),
    # indices = (1, None, '*', '(*)', '\\1')

    idx_str = indices
    # e.g., (1, None, '*', '(*)', '\\1')

    idx_str = ['w' if i == '*' else i for i in idx_str]
    # e..g, [1, None, 'w', '(*)', '\\1']

    idx_str = ['wp' if i == '(*)' else i for i in idx_str]
    # e.g., [1, None, 'w', 'wp', '\\1']

    idx_str = ['b{}'.format(i[1:]) if isinstance(i, str) and i.startswith('\\') else i for i in idx_str]
    # e.g., [1, None, 'w', 'wp', 'b1']

    idx_str = ['' if i is None else '{}{}'.format(idx_separator, i) for i in idx_str]
    # e.g., ['-1', '', '-w', '-wp', '-b1']

    colidxs = [n + i for n, i in zip(columnNames, idx_str)]
    # e.g., ['var1-1', 'var2', 'var3-w', 'var4-wp', 'var5-b1']

    middle = var_separator.join(colidxs)
    # e.g., 'var1-1.var2.var3-w.var4-wp.var5-b1'

    ret =  prefix + var_separator + middle + '.' + suffix
    # e.g., tbl_n_component.var1-1.var2.var3-w.var4-wp.var5-b1.txt

    return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest
import io

##__________________________________________________________________||
has_no_pandas = False
try:
    import pandas as pd
except ImportError:
    has_no_pandas = True

pytestmark = pytest.mark.skipif(has_no_pandas, reason="has no pandas")

if not has_no_pandas:
    from alphatwirl.collector import WritePandasDataFrameToFile

##__________________________________________________________________||
class MockOpen(object):
    def __init__(self, out):
        self._out = out

    def __call__(self, path):
        return self._out

def mockClose(file):
    pass

##__________________________________________________________________||
@pytest.fixture()
def out():
    return io.BytesIO()

@pytest.fixture()
def obj(out):
    ret = WritePandasDataFrameToFile("tbl.txt")
    ret._open = MockOpen(out)
    ret._close = mockClose
    yield ret

def test_repr(obj):
    repr(obj)

@pytest.mark.xfail(reason='fail with Pandas 0.20 because of its bug about the header indent')
def test_deliver(obj, out):
    results = pd.DataFrame(
        {
            'v1': [1, 2, 3],
            'n': [4.0, 3.0, 2.0],
            'nvar': [6.0, 9.0, 3.0],
        },
        columns = ['v1', 'n', 'nvar']
    )

    obj.deliver(results)

    expected = " v1  n  nvar\n  1  4     6\n  2  3     9\n  3  2     3\n".encode()
    assert expected == out.getvalue()

def test_deliver_empty_dataframe(obj, out):
    results = pd.DataFrame(
        {
            'v1': [ ],
            'n': [ ],
            'nvar': [ ],
        },
        columns = ['v1', 'n', 'nvar']
        )

    obj.deliver(results)

    expected = "v1 n nvar\n".encode()
    assert expected == out.getvalue()

def test_deliver_None_results(obj, out):
    obj.deliver(None)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl import progressbar

##__________________________________________________________________||
@pytest.fixture()
def moctime(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr(time, 'time', ret)
    ret.return_value = 1533374055.904203
    return ret

@pytest.fixture()
def obj(moctime):
    return progressbar.ProgressPrint()

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_report(obj, capsys):
    report = progressbar.ProgressReport('task1', 0, 10)
    obj.present(report)
    captured = capsys.readouterr()
    stdout_lines = captured.out.strip().split('\n')
    assert 2 == len(stdout_lines)
    assert '        0 /       10 (  0.00%) task1' == stdout_lines[1]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.loop import MPEventLoopRunner

##__________________________________________________________________||
@pytest.fixture()
def communicationChannel():
    return mock.Mock(name='communicationChannel')

@pytest.fixture()
def obj(communicationChannel):
    return MPEventLoopRunner(communicationChannel)

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_begin_end(obj, communicationChannel):
    obj.begin()
    communicationChannel.receive.return_value = [ ]
    obj.end()

def test_run(obj, communicationChannel):
    obj.begin()

    communicationChannel.put.side_effect = [0, 1]

    eventLoop1 = mock.Mock(name='eventLoop1')
    assert 0 == obj.run(eventLoop1)

    eventLoop2 = mock.Mock(name='eventLoop2')
    assert 1 == obj.run(eventLoop2)

    assert [mock.call(eventLoop1), mock.call(eventLoop2)] == communicationChannel.put.call_args_list

def test_run_multiple(obj, communicationChannel):
    obj.begin()
    communicationChannel.put_multiple.return_value = [0, 1]
    eventLoop1 = mock.Mock(name='eventLoop1')
    eventLoop2 = mock.Mock(name='eventLoop2')
    assert [0, 1] == obj.run_multiple([eventLoop1, eventLoop2])

    assert [mock.call([eventLoop1, eventLoop2])] == communicationChannel.put_multiple.call_args_list

def test_end(obj, communicationChannel):
    obj.begin()

    eventLoop1 = mock.Mock(name='eventLoop1')
    obj.run(eventLoop1)

    eventLoop2 = mock.Mock(name='eventLoop2')
    obj.run(eventLoop2)

    result1 = mock.Mock(name='result1')
    result2 = mock.Mock(name='result2')
    communicationChannel.receive.return_value = [result1, result2]
    assert [result1, result2] == obj.end()

def test_end_logging_wrong_number_or_results(obj, communicationChannel, caplog):
    obj.begin()

    eventLoop1 = mock.Mock(name='eventLoop1')
    obj.run(eventLoop1)

    eventLoop2 = mock.Mock(name='eventLoop2')
    obj.run(eventLoop2)

    result1 = mock.Mock(name='result1')
    result2 = mock.Mock(name='result2')
    communicationChannel.receive.return_value = [result1]

    with caplog.at_level(logging.WARNING, logger = 'alphatwirl'):
        results = obj.end()

    assert [result1] == results

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'MPEventLoopRunner' in caplog.records[0].name
    assert 'too few results received' in caplog.records[0].msg

##__________________________________________________________________||
def test_receive(obj, communicationChannel):
    obj.begin()

    communicationChannel.put.side_effect = [0, 1]

    eventLoop1 = mock.Mock(name='eventLoop1')
    assert 0 == obj.run(eventLoop1)

    eventLoop2 = mock.Mock(name='eventLoop2')
    assert 1 == obj.run(eventLoop2)

    result1 = mock.Mock(name='result1')
    result2 = mock.Mock(name='result2')
    communicationChannel.receive_all.return_value = [(0, result1), (1, result2)]
    assert [(0, result1), (1, result2)] == obj.receive()

def test_poll(obj, communicationChannel):
    obj.begin()

    communicationChannel.put.side_effect = [0, 1]

    eventLoop1 = mock.Mock(name='eventLoop1')
    assert 0 == obj.run(eventLoop1)

    eventLoop2 = mock.Mock(name='eventLoop2')
    assert 1 == obj.run(eventLoop2)

    result1 = mock.Mock(name='result1')
    result2 = mock.Mock(name='result2')
    communicationChannel.receive_finished.return_value = [(0, result1), (1, result2)]
    assert [(0, result1), (1, result2)] == obj.poll()

def test_receive_one(obj, communicationChannel):
    obj.begin()

    communicationChannel.put.side_effect = [0, 1]

    eventLoop1 = mock.Mock(name='eventLoop1')
    assert 0 == obj.run(eventLoop1)

    eventLoop2 = mock.Mock(name='eventLoop2')
    assert 1 == obj.run(eventLoop2)

    result1 = mock.Mock(name='result1')
    result2 = mock.Mock(name='result2')

    communicationChannel.receive_one.side_effect = [(0, result1), (1, result2), None]
    assert (0, result1) == obj.receive_one()
    assert (1, result2) == obj.receive_one()
    assert obj.receive_one() is None

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import BEvents as Events
    from alphatwirl.roottree import Branch

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def chain():
    input_file_names = [
        'sample_chain_01.root',
        'sample_chain_02.root',
        'sample_chain_03.root',
    ]
    input_paths = [
        os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', n)
        for n in input_file_names
    ]
    tree_name = 'tree'
    chain = ROOT.TChain(tree_name)
    for p in input_paths:
        chain.Add(p)
    yield chain

@pytest.fixture()
def events(chain):
    yield Events(chain)

def test_event(events):

    content_list = [
        # file 1
        [10, 20, 30],
        [24, 5],

        # file 2
        [3, 10],
        [5, 8, 32, 15, 2],
        [22, 11],

        # file 3
        [2, 7],
        [10, 100],
    ]

    for i, c in enumerate(content_list):
        event = events[i]
        assert len(c) == event.nvar[0]
        assert c == list(event.var)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

import collections
import itertools
import copy

from .convert import key_vals_dict_to_tuple_list

##__________________________________________________________________||
class Summarizer(object):
    def __init__(self, Summary):
        self._results = collections.defaultdict(Summary)
        self.Summary = Summary

    def __repr__(self):
        name_value_pairs = (
            ('Summary',  self.Summary),
            ('results', self._results),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def add(self, key, val=None, weight=1):
        self._results[key] += self.Summary(val, weight)

    def add_key(self, key):
        self._results[key]

    def keys(self):
        return self._results.keys()

    def __copy__(self):
        ret = self.__class__(self.Summary)
        self._add_results_inplace(ret._results, self._results)
        return ret

    def __add__(self, other):
        ret = copy.copy(self)
        if not other == 0: # other is 0 when e.g. sum([obj1, obj2])
            self._add_results_inplace(ret._results, other._results)
        return ret

    def __iadd__(self, other):
        self._add_results_inplace(self._results, other._results)
        return self

    def __radd__(self, other):
        return self.__add__(other)

    def _add_results_inplace(self, res1, res2):
        # res1 += res2, modify res1
        for k, v in res2.items():
            res1[k] += v

    def results(self):
        return self._results

    def to_key_vals_dict(self):
        keys_sorted = sorted(self._results.keys())
        ret = collections.OrderedDict([(k, self._results[k].contents) for k in keys_sorted])
        # e.g.,
        # OrderedDict([
        #     ((200, 2), [array([120, 240])]),
        #     ((300, 2), [array([490, 980])]),
        #     ((300, 3), [array([210, 420])])
        # ])
        return ret

    def to_tuple_list(self):
        key_vals_dict = self.to_key_vals_dict()
        ret = key_vals_dict_to_tuple_list(key_vals_dict, fill=0)
        # e.g.,
        # [
        #     (200, 2, 120, 240),
        #     (300, 2, 490, 980),
        #     (300, 3, 210, 420)
        # ]
        return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import shutil
import datetime
import tempfile
import imp
import logging
import tarfile
import gzip
import json
import re

try:
    import cPickle as pickle
except:
    import pickle

import alphatwirl

from alphatwirl.misc.deprecation import _renamed_class_method_option

##__________________________________________________________________||
class WorkingArea(object):
    """
        Args:
        topdir (str): a path to a directory in which a new directory will be created
        python_modules (list): names of python modules to be shipped to worker nodes

    """

    @_renamed_class_method_option(old='dir', new='topdir')
    def __init__(self, topdir, python_modules=()):
        self.topdir = topdir
        self.python_modules = python_modules
        self.path = None
        self.last_package_index = -1 # so it starts from 0

        self.executable = 'run.py'
        self.extra_input_files = set()

    def __repr__(self):
        name_value_pairs = (
            ('topdir', self.topdir),
            ('python_modules', self.python_modules),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def open(self):
        self.path = self._prepare_dir(self.topdir)
        self._copy_executable(area_path=self.path)
        self._save_logging_levels(area_path=self.path)
        self._put_python_modules(modules=self.python_modules, area_path=self.path)

    def put_package(self, package):

        self.last_package_index += 1
        package_index = self.last_package_index

        package_path = self.package_path(package_index)
        # relative to self.path, e.g., 'task_00009.p.gz'

        package_fullpath = os.path.join(self.path, package_path)
        # e.g., '{path}/tpd_20161129_122841_HnpcmF/task_00009.p.gz'

        with gzip.open(package_fullpath, 'wb') as f:
           pickle.dump(package, f, protocol=pickle.HIGHEST_PROTOCOL)
           f.close()

        return package_index

    def package_path(self, package_index):
        return 'task_{:05d}.p.gz'.format(package_index)

    def collect_result(self, package_index):

        dirname = 'task_{:05d}'.format(package_index)
        # e.g., 'task_00009'

        result_path = os.path.join(self.path, 'results', dirname, 'result.p.gz')
        # e.g., '{path}/tpd_20161129_122841_HnpcmF/results/task_00009/result.p.gz'

        try:
           with gzip.open(result_path, 'rb') as f:
              result = pickle.load(f)
        except (IOError, EOFError) as e:
           logger = logging.getLogger(__name__)
           logger.warning(e)
           return None

        return result

    def close(self):
        self.path = None

    def _prepare_dir(self, dir):

        alphatwirl.mkdir_p(dir)

        prefix = 'tpd_{:%Y%m%d_%H%M%S}_'.format(datetime.datetime.now())
        # e.g., 'tpd_20161129_122841_'

        path = tempfile.mkdtemp(prefix=prefix, dir=dir)
        # e.g., '{path}/tpd_20161129_122841_HnpcmF'

        return path

    def _copy_executable(self, area_path):
        thisdir = os.path.dirname(__file__)
        src = os.path.join(thisdir, self.executable)
        shutil.copy(src, area_path)

    def _save_logging_levels(self, area_path):
        logger_names = logging.Logger.manager.loggerDict.keys()
        loglevel_dict = {l: logging.getLogger(l).getEffectiveLevel() for l in logger_names}

        filename = 'logging_levels.json.gz'
        path = os.path.join(area_path, filename)

        json_str = json.dumps(loglevel_dict, indent=4, sort_keys=True)
        json_str = re.sub(r' *\n', '\n', json_str, flags=re.MULTILINE)
        json_str += "\n"
        json_bytes = json_str.encode('utf-8')
        with gzip.open(path, "w") as f:
            f.write(json_bytes)

        self.extra_input_files.add(filename)

    def _put_python_modules(self, modules, area_path):

        if not modules: return

        filename = 'python_modules.tar.gz'
        path = os.path.join(area_path, filename)
        tar = tarfile.open(path, 'w:gz')

        def tar_filter(tarinfo):
            exclude_extensions = ('.pyc', )
            exclude_names = ('.git', )
            if os.path.splitext(tarinfo.name)[1] in exclude_extensions: return None
            if os.path.basename(tarinfo.name) in exclude_names: return None
            return tarinfo

        for module in modules:
            imp_tuple = imp.find_module(module)
            path = imp_tuple[1]
            arcname = os.path.join('python_modules', module + imp_tuple[2][0])
            tar.add(path, arcname=arcname, filter=tar_filter)
        tar.close()

        self.extra_input_files.add(filename)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from .mock import MockReader, MockSummarizer

from alphatwirl.collector import ToTupleList

##__________________________________________________________________||
@pytest.fixture()
def obj():
    return ToTupleList(
        summaryColumnNames=('htbin', 'njetbin', 'n', 'nvar')
    )

def test_repr(obj):
    repr(obj)

def test_example(obj):
    reader1 = MockReader(
        MockSummarizer(
            [
                (200, 2, 120, 240),
            ]))

    reader2 = MockReader(
        MockSummarizer(
            [
                (300, 2, 490, 980),
                (300, 3, 210, 420),
            ]))

    reader3 = MockReader(
        MockSummarizer(
            [
                (300, 2, 20, 40),
                (300, 3, 15, 30),
            ]))

    reader4 = MockReader(MockSummarizer([]))

    dataset_readers_list = [
        ('QCD', (reader1, reader2)),
        ('TTJets', (reader3, )),
        ('WJets', (reader4, )),
        ('ZJets', ( )),
    ]

    expected = [
        ('htbin', 'njetbin', 'n', 'nvar'),
        (200, 2, 120, 240),
        (300, 2, 490, 980),
        (300, 3, 210, 420),
        (300, 2, 20, 40),
        (300, 3, 15, 30)
    ]

    actual = obj.combine(dataset_readers_list)

    assert expected == actual

def test_combine_oneReader(obj):

    reader1 = MockReader(
        MockSummarizer(
            [
                (200, 2, 120, 240),
                (300, 2, 310, 620),
            ]))

    dataset_readers_list = [
        ('QCD', (reader1, )),
    ]

    expected = [
        ('htbin', 'njetbin', 'n', 'nvar'),
        (200, 2, 120, 240),
        (300, 2, 310, 620),
    ]

    actual = obj.combine(dataset_readers_list)

    assert expected == actual

def test_combine_all_empty_contents(obj):
    reader1 = MockReader(
        MockSummarizer(
            [ ]))

    reader2 = MockReader(
        MockSummarizer(
            [ ]))

    reader3 = MockReader(
        MockSummarizer(
            [ ]))

    dataset_readers_list = [
        ('QCD', (reader1, reader2)),
        ('TTJets', (reader3, )),
    ]

    expected = [
        ('htbin', 'njetbin', 'n', 'nvar'),
    ]

    actual = obj.combine(dataset_readers_list)

    assert expected == actual

def test_combine_empty_pairs(obj):
    dataset_readers_list = [ ]

    expected = None

    actual = obj.combine(dataset_readers_list)

    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from alphatwirl.binning.search import linear_search, binary_search

##__________________________________________________________________||
@pytest.mark.parametrize('search', [linear_search, binary_search])
@pytest.mark.parametrize('val, boundaries, expected', [
    pytest.param(0.1, [0.1], 0, id='one-boundary'),
    pytest.param(0.1, [0.1, 0.2], 0, id='two-boundaries-at-lowedge'),
    pytest.param(0.15, [0.1, 0.2], 0, id='two-boundaries-in-bin'),
    pytest.param(0.2, [0.1, 0.2], 1, id='two-boundaries-at-upedge'),
    pytest.param(0.1, [0.1, 0.2, 0.3], 0, id='three-boundaries'),
    pytest.param(0.15, [0.1, 0.2, 0.3], 0, id='three-boundaries'),
    pytest.param(0.2, [0.1, 0.2, 0.3], 1, id='three-boundaries'),
    pytest.param(0.25, [0.1, 0.2, 0.3], 1, id='three-boundaries'),
    pytest.param(0.3, [0.1, 0.2, 0.3], 2, id='three-boundaries'),
    pytest.param(0.1, [0.1, 0.2, 0.3, 0.4], 0, id='four-boundaries'),
    pytest.param(0.15, [0.1, 0.2, 0.3, 0.4], 0, id='four-boundaries'),
    pytest.param(0.2, [0.1, 0.2, 0.3, 0.4], 1, id='four-boundaries'),
    pytest.param(0.25, [0.1, 0.2, 0.3, 0.4], 1, id='four-boundaries'),
    pytest.param(0.3, [0.1, 0.2, 0.3, 0.4], 2, id='four-boundaries'),
    pytest.param(0.35, [0.1, 0.2, 0.3, 0.4], 2, id='four-boundaries'),
    pytest.param(0.4, [0.1, 0.2, 0.3, 0.4], 3, id='four-boundaries'),
])
def test_search(search, val, boundaries, expected):
    assert expected == search(val, boundaries)

##__________________________________________________________________||



import unittest
import collections

from alphatwirl.loop import CollectorDelegate

##__________________________________________________________________||
class MockCollector(object):
    def __init__(self, ret = None):
        self.ret = ret

    def collect(self, dataset_readers_list):
        self.args = dataset_readers_list
        return self.ret

##__________________________________________________________________||
MockResult = collections.namedtuple('MockResult', 'name')

##__________________________________________________________________||
MockArgs = collections.namedtuple('MockArgs', 'name')

##__________________________________________________________________||
class TestCollectorDelegate(unittest.TestCase):

    def setUp(self):
        """
        1:delegate - 2:collector

        """
        self.result = MockResult('result')
        self.collector = MockCollector(ret = self.result)
        self.obj = CollectorDelegate(self.collector)

    def test_repr(self):
        repr(self.obj)

    def test_collect(self):

        args = MockArgs('args')
        self.assertIs(self.result, self.obj.collect(args))
        self.assertIs(args, self.collector.args)

##__________________________________________________________________||




# Tai Sakuma <tai.sakuma@gmail.com>
from ..progressbar import ProgressReport

##__________________________________________________________________||
class EventLoopProgressReportWriter(object):
    """A progress report writer of an event loop

    """
    def __repr__(self):
        return '{}()'.format(
            self.__class__.__name__
        )

    def write(self, taskid, config, event):
        return ProgressReport(
            name = config.name,
            done = event.iEvent + 1,
            total = event.nEvents,
            taskid = taskid
        )

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class ToTupleListWithDatasetColumn(object):
    def __init__(self, summaryColumnNames,
                 datasetColumnName='component'
                 ):

        self.summaryColumnNames = summaryColumnNames
        self.datasetColumnName = datasetColumnName

        self._repr_pairs = [
            ('summaryColumnNames', self.summaryColumnNames),
            ('datasetColumnName',  self.datasetColumnName),
        ]

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in self._repr_pairs]),
        )

    def combine(self, dataset_readers_list):


        if len(dataset_readers_list) == 0: return None

        # e.g.,
        # dataset_readers_list = [
        #     ('QCD',    (reader1, reader2)),
        #     ('TTJets', (reader3, )),
        #     ('WJets',  (reader4, )),
        #     ('ZJets',  ( )),
        # ]

        # remove entries with no readers
        dataset_readers_list = [l for l in dataset_readers_list if l[1]]
        if len(dataset_readers_list) == 0: return None

        dataset_summarizers_list = [(d, tuple(r.results() for r in rs)) for d, rs in dataset_readers_list]
        # e.g.,
        # dataset_summarizers_list = [
        #     ('QCD',    (summarizer1, summarizer2)),
        #     ('TTJets', (summarizer3, ),
        #     ('WJets',  (summarizer4, ),
        # ]

        dataset_summarizer_pairs = [(d, sum(s)) for d, s in  dataset_summarizers_list]
        # e.g.,
        # dataset_summarizer_pairs = [
        #     ('QCD',    summarizer1 + summarizer2),
        #     ('TTJets', summarizer3),
        #     ('WJets',  summarizer4),
        # ]
        # note: summarizers can be added

        dataset_tuple_list_pairs = [(d, s.to_tuple_list()) for d, s in dataset_summarizer_pairs]
        # e.g.,
        # dataset_tuple_list_pairs = [
        #     ('QCD', [
        #         (200, 2, 120, 240),
        #         (300, 2, 490, 980),
        #         (300, 3, 210, 420)
        #     ]),
        #     ('TTJets', [
        #         (300, 2, 20, 40),
        #         (300, 3, 15, 30)
        #     ]),
        #     ('WJets', [])
        # ]

        ret = [ ]
        for dataset, tuple_list in dataset_tuple_list_pairs:
            ret.extend([(dataset, ) + e for e in tuple_list])
        # e.g.,
        # [
        #     ('QCD',    200, 2, 120, 240),
        #     ('QCD',    300, 2, 490, 980),
        #     ('QCD',    300, 3, 210, 420),
        #     ('TTJets', 300, 2,  20,  40),
        #     ('TTJets', 300, 3,  15,  30)
        # ]

        header = (self.datasetColumnName, ) + self.summaryColumnNames

        ret.insert(0, header)
        # e.g.,
        # [
        #     ('dataset', 'htbin', 'njetbin', 'n', 'nvar'),
        #     ('QCD',         200,         2, 120,    240),
        #     ('QCD',         300,         2, 490,    980),
        #     ('QCD',         300,         3, 210,    420),
        #     ('TTJets',      300,         2,  20,     40),
        #     ('TTJets',      300,         3,  15,     30)
        # ]

        return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class CommunicationChannel0(object):
    """A communication channel for the single process mode

    An alternative to `CommunicationChannel`. However, unlike
    `CommunicationChannel`, this class does not send tasks to workers.
    Instead, it directly executes tasks.

    This class has the same interface as the class
    `CommunicationChannel`. When this class is used as a substitute
    for `CommunicationChannel`, the tasks will be sequentially
    executed in the foreground.

    """

    def __init__(self):
        self.taskidx = -1 # so it starts from 0
        self.taskidx_result_pairs = [ ]

    def __repr__(self):
        return '{}()'.format(
            self.__class__.__name__,
        )

    def begin(self):
        pass

    def put(self, task, *args, **kwargs):
        self.taskidx += 1
        result = task(*args, **kwargs)
        self.taskidx_result_pairs.append((self.taskidx, result))
        return self.taskidx

    def put_multiple(self, task_args_kwargs_list):
        task_idxs = [ ]
        for t in task_args_kwargs_list:
            try:
                task = t['task']
                args = t.get('args', ())
                kwargs = t.get('kwargs', {})
                task_idx = self.put(task, *args, **kwargs)
            except TypeError:
                task_idx = self.put(t)
            task_idxs.append(task_idx)
        return task_idxs

    def receive_finished(self):
        return self.receive_all()

    def receive_one(self):
        if self.taskidx_result_pairs:
            return self.taskidx_result_pairs.pop(0)
        return None

    def receive_all(self):
        ret = self.taskidx_result_pairs[:]
        del self.taskidx_result_pairs[:]
        return ret

    def receive(self):
        ret = [r for _, r in self.receive_all()]
        return ret

    def terminate(self):
        pass

    def end(self):
        pass

##__________________________________________________________________||



from alphatwirl.binning import Combine, Round, RoundLog
import unittest

##__________________________________________________________________||
def plus2(val): return val + 2

##__________________________________________________________________||
class TestCombine(unittest.TestCase):
    def test_call(self):
        low = Round(20.0, 100)
        high = RoundLog(0.1, 100)
        binning = Combine(low =low, high = high, at = 100)
        self.assertEqual(0, binning(10))
        self.assertEqual(80, binning(90))
        self.assertEqual(100, binning(100))

        low = Round(10.0, 50)
        high = RoundLog(0.1, 50)
        binning = Combine(low =low, high = high, at = 50)
        self.assertEqual(10, binning(11))
        self.assertEqual(40, binning(40))
        self.assertAlmostEqual(50, binning(50))

    def test_next(self):
        low = Round(20.0, 100)
        high = RoundLog(0.1, 100)
        binning = Combine(low =low, high = high, at = 100)
        self.assertEqual(20, binning.next(10))
        self.assertEqual(100, binning.next(90))
        self.assertEqual(125.89254117941675, binning.next(100))

        low = Round(10.0, 50)
        high = RoundLog(0.1, 50)
        binning = Combine(low =low, high = high, at = 50)
        self.assertEqual(20, binning.next(10))
        self.assertAlmostEqual(50, binning.next(45))
        self.assertEqual(62.94627058970836, binning.next(50))

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

import ROOT

from alphatwirl.misc.deprecation import _deprecated

from .Events import Events
from .inspect import is_ROOT_null_pointer

##__________________________________________________________________||
@_deprecated(msg='use BuildEvents instead')
class EventBuilder(object):
    def __init__(self, config, EventsClass=Events):
        self.config = config
        self.EventsClass = EventsClass

    def __repr__(self):
        return '{}({!r})'.format(
            self.__class__.__name__,
            self.config
        )

    def __call__(self):
        chain = ROOT.TChain(self.config.treeName)
        for path in self.config.inputPaths:
            file_ = ROOT.TFile.Open(path)
            if is_ROOT_null_pointer(file_) or file_.IsZombie():
                logger = logging.getLogger(__name__)
                logger.error('cannot open {}'.format(path))
                raise OSError('cannot open {}'.format(path))
            chain.Add(path)
        events = self.EventsClass(chain, self.config.maxEvents, self.config.start)
        events.config = self.config
        return events

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import sys
import copy
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

import alphatwirl

class A(object):
    def __init__(self, a):
        pass

class B(object):
    def __init__(self):
        self._repr = 'B'

    def __repr__(self):
        return self._repr

## In Python 2 with mock 2.0.0, mock.Mock with spec sometimes cannot
## be copied or deep-copied.
## The tests in this file demonstrate.

##__________________________________________________________________||
@pytest.mark.skipif(sys.version_info < (3, 0), reason='require python 3')
def test_copy_a():
    obj = mock.Mock(spec=A)
    copy.deepcopy(obj) # doesn't work in python 2 with mock 2.0.0

@pytest.mark.skipif(sys.version_info < (3, 0), reason='require python 3')
def test_copy_b():
    obj = mock.Mock(spec=B)
    obj_copy = copy.copy(obj)
    repr(obj_copy) # doesn't work in python 2 with mock 2.0.0

##__________________________________________________________________||




# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class Branch(object):
    """This class encloses an array.array object, which is typically used
    to set the address of a branch of a ROOT TTree. This class is
    useful for fast access to contents of TTree.

    While TTree can hold many types of data, including objects of user
    defined classes, this class can be used only for objects of simple
    types, such as Int_t, Double_t, and arrays of simple types.


    Examples
    --------
    Suppose jet_pt is an instance of this class and the corresponding
    object in TTree is an array.

    >>> len(jet_pt)
    3

    >>> [v for v in jet_pt]
    [127.16558074951172, 68.16969299316406, 53.75463104248047]

    When the corresponding object in TTree is not an array but a
    simple type, the length is 1 and the only element is the value.

    >>> len(mht)
    1

    >>> mht[0]
    73.7677


    """


    def __init__(self, name, array, countarray):
        self.name = name
        self.array = array
        self.countarray = countarray

    def __repr__(self):
        return '{}(name={!r}, array={!r}, countarray={!r})'.format(
            self.__class__.__name__,
            self.name,
            self.array,
            self.countarray
        )

    def __getitem__(self, i):
        if self.countarray is None and 0 != i:
            raise IndexError('the index should be zero for this branch: {}[{}]'.format(self.name, i))
        if self.countarray is not None and i >= self.countarray[0]:
            raise IndexError('the index is out of range: {}[{}]'.format(self.name, i))
        return self.array[i]

    def __len__(self):
        if self.countarray is None: return 1
        return self.countarray[0]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

from alphatwirl.progressbar import ProgressReport

##__________________________________________________________________||
def test_repr():
    obj = ProgressReport(name = 'dataset1', done = 124, total = 1552, taskid = 1212)
    repr(obj)

def test_init():
    obj = ProgressReport(name = 'dataset1', done = 124, total = 1552, taskid = 1212)

def test_init_no_taskid():
    obj = ProgressReport(name = 'dataset1', done = 124, total = 1552)
    assert obj.taskid == 'dataset1'

def test_last():
    obj = ProgressReport(name = 'dataset1', done = 124, total = 1552)
    assert not obj.last()

    obj = ProgressReport(name = 'dataset1', done = 1552, total = 1552)
    assert obj.last()

    obj = ProgressReport(name = 'dataset1', done = 0, total = 0)
    assert obj.last()

def test_first():
    obj = ProgressReport(name = 'dataset1', done = 124, total = 1552)
    assert not obj.last()

    obj = ProgressReport(name = 'dataset1', done = 0, total = 1552)
    assert obj.first()

    obj = ProgressReport(name = 'dataset1', done = 0, total = 0)
    assert obj.first()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class CollectorDelegate(object):

    """
    """

    def __init__(self, collector):
        self.collector = collector

    def __repr__(self):
        name_value_pairs = (
            ('collector', self.collector),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{} = {!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def collect(self, dataset_readers_list):
        return self.collector.collect(dataset_readers_list)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

##__________________________________________________________________||
has_no_pandas = False
try:
    import pandas as pd
except ImportError:
    has_no_pandas = True

pytestmark = pytest.mark.skipif(has_no_pandas, reason="has no pandas")

from .mock import MockReader, MockSummarizer

if not has_no_pandas:
    from pandas.util.testing import assert_frame_equal
    from alphatwirl.collector import ToDataFrameWithDatasetColumn

##__________________________________________________________________||
@pytest.fixture()
def obj():
    return ToDataFrameWithDatasetColumn(
        summaryColumnNames=('htbin', 'njetbin', 'n', 'nvar'),
        datasetColumnName = 'dataset'
    )

def test_repr(obj):
    repr(obj)

def test_example(obj):
    reader1 = MockReader(
        MockSummarizer(
            [
                (200, 2, 120, 240),
            ]))

    reader2 = MockReader(
        MockSummarizer(
            [
                (300, 2, 490, 980),
                (300, 3, 210, 420),
            ]))

    reader3 = MockReader(
        MockSummarizer(
            [
                (300, 2, 20, 40),
                (300, 3, 15, 30),
            ]))

    reader4 = MockReader(MockSummarizer([]))

    dataset_readers_list = [
        ('QCD', (reader1, reader2)),
        ('TTJets', (reader3, )),
        ('WJets', (reader4, )),
        ('ZJets', ( )),
    ]

    expected = pd.DataFrame(
        {
            'dataset': ['QCD', 'QCD', 'QCD', 'TTJets', 'TTJets'],
            'htbin': [200, 300, 300, 300, 300],
            'njetbin': [2, 2, 3, 2, 3],
            'n': [120, 490, 210, 20, 15],
            'nvar': [240, 980, 420, 40, 30],
        })

    actual = obj.combine(dataset_readers_list)
    assert_frame_equal(expected.sort_index(axis = 1), actual.sort_index(axis = 1), check_names=True)

def test_combine_all_empty_counts(obj):
    reader1 = MockReader(MockSummarizer([]))
    reader2 = MockReader(MockSummarizer([]))

    dataset_readers_list = [
        ('data1', (reader1, )),
        ('data2', (reader2, )),
    ]

    # # this will make the index Index([], dtype='object')
    # expected = pd.DataFrame(
    #     {
    #         'dataset': [ ],
    #         'htbin': [ ],
    #         'njetbin': [ ],
    #         'n': [ ],
    #         'nvar': [ ],
    #         }
    #     )

    # this will make the index Int64Index([], dtype='int64')
    expected = pd.DataFrame(
         [],
         columns = ('dataset', 'htbin', 'njetbin', 'n', 'nvar')
    )

    actual = obj.combine(dataset_readers_list)
    assert_frame_equal(expected.sort_index(axis = 1), actual.sort_index(axis = 1), check_names=True)

def test_combine_empty_pairs(obj):

    dataset_readers_list = [
    ]

    actual = obj.combine(dataset_readers_list)
    assert actual is None

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from alphatwirl.selection.factories.expand import expand_path_cfg

from alphatwirl.selection.factories.factory import FactoryDispatcher
from alphatwirl.selection.modules.LambdaStr import LambdaStr
from alphatwirl.selection.modules import All, Any, Not

##__________________________________________________________________||
@pytest.fixture()
def alias_dict():
    return {
        'alias1': 'ev : ev.var1[0] >= 10',
        'alias2': ('ev : ev.var2[0] >= 20', dict(name='name2')),
        'alias3': 'alias1',
        'alias4': 'alias3',
        'alias5': 'ev : ev.var4[0] == {n}',
        'alias6': ('ev : {low} <= ev.var5[0] < {high}', dict(low=11, high=20))
    }

##__________________________________________________________________||
# path_cfg, expanded, obj
params = [
    pytest.param(
        'alias1',
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : ev.var1[0] >= 10',
            name='alias1'
        ),
        LambdaStr(
            name='alias1',
            lambda_str='ev : ev.var1[0] >= 10'
        ),
        id='alias1'
    ),
    pytest.param(
        ('alias1', dict(name='name1')),
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : ev.var1[0] >= 10',
            name='name1'
        ),
        LambdaStr(name='name1', lambda_str='ev : ev.var1[0] >= 10'),
        id='alias1:with-name'
    ),
    pytest.param(
        'alias2',
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : ev.var2[0] >= 20',
            name='name2' #  name has priority over alias
        ),
        LambdaStr(name='name2', lambda_str='ev : ev.var2[0] >= 20'),
        id='alias2:name-priority-over-alias'
    ),
    pytest.param(
        ('alias2', dict(name='new_name2')),
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : ev.var2[0] >= 20',
            name='new_name2' # name can be overridden
        ),
        LambdaStr(name='new_name2', lambda_str='ev : ev.var2[0] >= 20'),
        id='alias2:name-overridden'
    ),
    pytest.param(
        'alias3',
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : ev.var1[0] >= 10',
            name='alias3' # the outermost alias has priority
        ),
        LambdaStr(name='alias3', lambda_str='ev : ev.var1[0] >= 10'),
        id='alias3:alias-of-alias'
    ),
    pytest.param(
        'alias4',
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : ev.var1[0] >= 10',
            name='alias4' # the outermost alias has priority
        ),
        LambdaStr(name='alias4', lambda_str='ev : ev.var1[0] >= 10'),
        id='alias4:alias-of-alias-of-alias'
    ),
    pytest.param(
        ('alias5', dict(n=30)),
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : ev.var4[0] == {n}', # not formatted
            n=30,
            name='alias5'
        ),
        LambdaStr(name='alias5', lambda_str='ev : ev.var4[0] == 30'),
        id='alias5:not-formatted'
    ),
    pytest.param(
        'alias6',
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : {low} <= ev.var5[0] < {high}',
            low=11,
            high=20,
            name='alias6',
        ),
        LambdaStr(name='alias6', lambda_str='ev : 11 <= ev.var5[0] < 20'),
        id='alias6:not-formatted-with-default-values'
    ),
    pytest.param(
        ('alias6', dict(high=30)),
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : {low} <= ev.var5[0] < {high}',
            low=11,
            high=30,
            name='alias6'
        ),
        LambdaStr(name='alias6', lambda_str='ev : 11 <= ev.var5[0] < 30'),
        id='alias6:not-formatted-with-default-values-overridden'
    ),
]

##__________________________________________________________________||
@pytest.mark.parametrize('path_cfg, expected, _', params)
def test_alias(alias_dict, path_cfg, expected, _):
    actual = expand_path_cfg(path_cfg=path_cfg, alias_dict=alias_dict)
    assert expected == actual

    # give expanded one
    actual = expand_path_cfg(path_cfg=actual, alias_dict=alias_dict)
    assert expected == actual


@pytest.mark.parametrize('path_cfg, expected, _', params)
def test_nested(alias_dict, path_cfg, expected, _):
    path_cfg = dict(All=(path_cfg, ))
    actual = expand_path_cfg(path_cfg=path_cfg, alias_dict=alias_dict)
    expected = dict(factory='AllFactory', components=(expected, ))
    assert expected == actual

    path_cfg = dict(All=(path_cfg, ))
    actual = expand_path_cfg(path_cfg=path_cfg, alias_dict=alias_dict)
    expected = dict(factory='AllFactory', components=(expected, ))
    assert expected == actual

    path_cfg = dict(Not=path_cfg)
    actual = expand_path_cfg(path_cfg=path_cfg, alias_dict=alias_dict)
    expected = dict(factory='NotFactory', components=(expected, ))
    assert expected == actual

    path_cfg = dict(Any=(path_cfg, ))
    actual = expand_path_cfg(path_cfg=path_cfg, alias_dict=alias_dict)
    expected = dict(factory='AnyFactory', components=(expected, ))
    assert expected == actual

    path_cfg = dict(Any=(path_cfg, ))
    actual = expand_path_cfg(path_cfg=path_cfg, alias_dict=alias_dict)
    expected = dict(factory='AnyFactory', components=(expected, ))
    assert expected == actual

    path_cfg = dict(Not=path_cfg)
    actual = expand_path_cfg(path_cfg=path_cfg, alias_dict=alias_dict)
    expected = dict(factory='NotFactory', components=(expected, ))
    assert expected == actual

    # give expanded one
    actual = expand_path_cfg(path_cfg=actual, alias_dict=alias_dict)
    assert expected == actual

##__________________________________________________________________||
@pytest.mark.parametrize('path_cfg, _, expected', params)
def test_factory(alias_dict, path_cfg, _, expected):

    kargs = dict(
        AllClass=All, AnyClass=Any, NotClass=Not,
        LambdaStrClass=LambdaStr, aliasDict=alias_dict,
    )

    obj = FactoryDispatcher(path_cfg=path_cfg, **kargs)
    assert repr(expected) == repr(obj)
    assert str(expected) == str(obj)

@pytest.mark.parametrize('path_cfg, _, expected', params)
def test_factory_nested(alias_dict, path_cfg, _, expected):

    kargs = dict(
        AllClass=All, AnyClass=Any, NotClass=Not,
        LambdaStrClass=LambdaStr, aliasDict=alias_dict,
    )

    path_cfg = dict(All=(path_cfg, ))
    expected = All(name='All', selections=[expected])

    path_cfg = dict(All=(path_cfg, ))
    expected = All(name='All', selections=[expected])

    path_cfg = dict(Not=path_cfg)
    expected = Not(name='Not', selection=expected)

    path_cfg = dict(Any=(path_cfg, ))
    expected = Any(name='Any', selections=[expected])

    path_cfg = dict(Any=(path_cfg, ))
    expected = Any(name='Any', selections=[expected])

    path_cfg = dict(Not=path_cfg)
    expected = Not(name='Not', selection=expected)

    obj = FactoryDispatcher(path_cfg=path_cfg, **kargs)
    assert repr(expected) == repr(obj)
    assert str(expected) == str(obj)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import copy
import collections
import logging

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.loop import CollectorComposite, ReaderComposite

##__________________________________________________________________||
class MockCollector(object):
    def __init__(self, ret=None):
        self.collected = None
        self.ret = ret

    def collect(self, dataset_readers_list):
        self.collected = dataset_readers_list
        return self.ret

##__________________________________________________________________||
class MockReader(object):
    pass

##__________________________________________________________________||
class MockDataset(object):
    pass

##__________________________________________________________________||
def test_repr():
    obj = CollectorComposite()
    repr(obj)

def test_deprecated_option(caplog):
    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        CollectorComposite(progressReporter=None)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'CollectorComposite' in caplog.records[0].name
    assert 'deprecated' in caplog.records[0].msg

def test_collect():
    """
    1:composite
        |- 2:composite
        |        |- 3:leaf
        |        |- 4:leaf
        |
        |- 5:leaf
    """

    ## build collector
    collector1 = CollectorComposite()
    collector2 = CollectorComposite()
    collector3 = MockCollector('result3')
    collector4 = MockCollector('result4')
    collector5 = MockCollector('result5')

    collector1.add(collector2)
    collector2.add(collector3)
    collector2.add(collector4)
    collector1.add(collector5)

    ## build reader
    reader1 = ReaderComposite()
    reader2 = ReaderComposite()
    reader3 = MockReader()
    reader4 = MockReader()
    reader5 = MockReader()

    reader1.add(reader2)
    reader2.add(reader3)
    reader2.add(reader4)
    reader1.add(reader5)

    ## copy readers
    reader1_copy1 = copy.deepcopy(reader1)
    reader1_copy2 = copy.deepcopy(reader1)
    reader1_copy3 = copy.deepcopy(reader1)
    reader1_copy4 = copy.deepcopy(reader1)

    ## build data set
    dataset1 = MockDataset()
    dataset2 = MockDataset()
    dataset3 = MockDataset()

    assert collector3.collected is None
    assert collector4.collected is None
    assert collector5.collected is None

    dataset_readers_list = [
        (dataset1, (reader1_copy1, reader1_copy2, reader1_copy3)),
        (dataset2, ( )),
        (dataset3, (reader1_copy4, )),
    ]

    assert [
        ['result3', 'result4'],
        'result5'
    ] == collector1.collect(dataset_readers_list)

    assert [
        (dataset1, (reader1_copy1.readers[0].readers[0], reader1_copy2.readers[0].readers[0], reader1_copy3.readers[0].readers[0])),
        (dataset2, ( )),
        (dataset3, (reader1_copy4.readers[0].readers[0], )),
    ] == collector3.collected

    assert [
        (dataset1, (reader1_copy1.readers[0].readers[1], reader1_copy2.readers[0].readers[1], reader1_copy3.readers[0].readers[1])),
        (dataset2, ( )),
        (dataset3, (reader1_copy4.readers[0].readers[1], )),
    ] == collector4.collected

    assert [
        (dataset1, (reader1_copy1.readers[1], reader1_copy2.readers[1], reader1_copy3.readers[1])),
        (dataset2, ( )),
        (dataset3, (reader1_copy4.readers[1], )),
    ] == collector5.collected

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

from alphatwirl import listToAlignedText

##__________________________________________________________________||
def test_deprecated(caplog):
    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        listToAlignedText('')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'listToAlignedText' in caplog.records[0].name
    assert 'deprecated' in caplog.records[0].msg

def test_name():
    assert  'listToAlignedText' == listToAlignedText.__name__

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import math
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.summary import KeyValueComposer

##__________________________________________________________________||
class MockEvent(object):
    pass

class MockBinningEcho(object):
    def __call__(self, val):
        return val

class MockBinningFloor(object):
    def __init__(self, max=None):
        self.max = max

    def __call__(self, val):
        if val is None:
            return None
        if self.max is not None and not val <= self.max:
            return None
        return int(math.floor(val))

##__________________________________________________________________||
class MockArrayReader(object):
    def __init__(self, arrays=None, idxs_conf=None, backref_idxs=None):
        self.ret = ( )
        pass

    def read(self):
        return self.ret

class MockArrayReaderRaise(object):
    def __init__(self, arrays=None, idxs_conf=None, backref_idxs=None):
        pass

    def read(self):
        raise Exception('raised by MockArrayReaderRaise')

##__________________________________________________________________||
def test_repr():
    obj = KeyValueComposer()
    repr(obj)

##__________________________________________________________________||
def test_init_raise_wrong_key_length():
    with pytest.raises(ValueError):
        KeyValueComposer(
            keyAttrNames=('var1', ),
            keyIndices=(0, 1)
        )

def test_init_raise_wrong_val_length():
    with pytest.raises(ValueError):
        KeyValueComposer(
            valAttrNames = ('var1', ),
            valIndices = (0, 1)
        )

def test_init_raise_wrong_binning_length():
    with pytest.raises(ValueError):
        KeyValueComposer(
            keyAttrNames = ('var1', ),
            binnings = (MockBinningEcho(), MockBinningEcho()),
        )

##__________________________________________________________________||
def test_collect_arrays():
    obj = KeyValueComposer()

    event = MockEvent()
    event.var1 = [ ]
    event.var2 = [ ]
    event.var3 = [ ]
    attr_names = ('var1', 'var2', 'var3')
    arrays = obj._collect_arrays(event, attr_names)
    assert event.var1 is arrays[0]
    assert event.var2 is arrays[1]
    assert event.var3 is arrays[2]

def test_collect_arrays_error(caplog):
    obj = KeyValueComposer()

    event = MockEvent()
    event.var1 = [ ]
    event.var2 = [ ]
    attr_names = ('var1', 'var2', 'var3') # var3 doesn't exist

    with caplog.at_level(logging.WARNING):
        arrays = obj._collect_arrays(event, attr_names)

    assert arrays is None

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'KeyValueComposer' in caplog.records[0].name
    assert 'has no attribute' in caplog.records[0].msg

##__________________________________________________________________||
def test_call_inactive():
    obj = KeyValueComposer()
    obj.ArrayReader = MockArrayReader
    obj.active = False

    event = MockEvent()
    assert () == obj(event)

def test_call_raise():
    obj = KeyValueComposer()
    obj.active = True

    array_reader = MockArrayReaderRaise()
    obj._array_reader = array_reader

    event = MockEvent()
    with pytest.raises(Exception):
        obj(event)

##__________________________________________________________________||
def test_call_NoneKey_NoneVal():
    obj = KeyValueComposer()
    obj.active = True

    obj._lenkey = 0
    obj.binnings = None

    array_reader = MockArrayReader()
    obj._array_reader = array_reader

    event = MockEvent()

    array_reader.ret = ((), )
    assert (((), ()), ) == obj(event)

def test_call_1Key_NoneVal():
    obj = KeyValueComposer()
    obj.active = True

    obj._lenkey = 1
    obj.binnings = (MockBinningFloor(max=30), )

    array_reader = MockArrayReader()
    obj._array_reader = array_reader

    event = MockEvent()

    array_reader.ret = ((12.5, ), )
    assert (
        ((12,), ()),
    ) == obj(event)

    array_reader.ret = ((32.5, ), ) # out of range
    assert ( ) == obj(event)

def test_call_NoneKey_1Val():
    obj = KeyValueComposer()
    obj.active = True

    obj._lenkey = 0
    obj.binnings = None

    array_reader = MockArrayReader()
    obj._array_reader = array_reader

    event = MockEvent()

    array_reader.ret = ((12.8, ), )
    assert (
        ((), (12.8, )),
    ) == obj(event)

def test_call_1Key_1Val():
    obj = KeyValueComposer()
    obj.active = True

    obj._lenkey = 1
    obj.binnings = (MockBinningFloor(max=30), )

    array_reader = MockArrayReader()
    obj._array_reader = array_reader

    event = MockEvent()

    array_reader.ret = ((12.5, 20.3), )
    assert (
        ((12, ), (20.3, )),
    ) == obj(event)

    array_reader.ret = ((32.5, 20.3), )
    assert () == obj(event)

def test_call_2Keys_NoneVal():
    obj = KeyValueComposer()
    obj.active = True

    obj._lenkey = 2
    obj.binnings = (MockBinningFloor(max=30), MockBinningFloor(max=50))

    array_reader = MockArrayReader()
    obj._array_reader = array_reader

    event = MockEvent()

    array_reader.ret = ((15.3, 22.8), )
    assert (
        ((15, 22), ()),
    ) == obj(event)

    array_reader.ret = ((45.3, 22.8), ) # 1st element out of range
    assert () == obj(event)

    array_reader.ret = ((15.3, 52.8), ) # 2nd element out of range
    assert () == obj(event)

    array_reader.ret = ((45.3, 52.8), ) # both out of range
    assert () == obj(event)

def test_call_2Keys_1Val():
    obj = KeyValueComposer()
    obj.active = True

    obj._lenkey = 2
    obj.binnings = (MockBinningFloor(max=30), MockBinningFloor(max=50))

    array_reader = MockArrayReader()
    obj._array_reader = array_reader

    event = MockEvent()

    array_reader.ret = ((15.3, 22.8, 101.1), )
    assert (
        ((15, 22), (101.1, )),
    ) == obj(event)

    array_reader.ret = ((45.3, 22.8, 101.1), ) # 1st element out of range
    assert () == obj(event)

    array_reader.ret = ((15.3, 52.8, 101.1), ) # 2nd element out of range
    assert () == obj(event)

    array_reader.ret = ((45.3, 52.8, 101.1), ) # both out of range
    assert () == obj(event)

def test_call_2Keys_1Val_NoneBinnings():
    obj = KeyValueComposer()
    obj.active = True

    obj._lenkey = 2
    obj.binnings = None

    array_reader = MockArrayReader()
    obj._array_reader = array_reader

    event = MockEvent()

    array_reader.ret = ((15.3, 22.8, 101.1), )
    assert (
        ((15.3, 22.8), (101.1, )),
    ) == obj(event)

##__________________________________________________________________||
def test_example_back_reference_twice():
    obj = KeyValueComposer(
        keyAttrNames=('ev', 'jet_pt', 'jet_eta', 'mu_pt', 'mu_eta', 'jet_phi'),
        binnings=(
            MockBinningFloor(),
            MockBinningFloor(),
            MockBinningFloor(max=3), # <- use max for jet_eta
            MockBinningFloor(),
            MockBinningFloor(max=2), # <- use max for mu_eta
            MockBinningEcho(),
        ),
        keyIndices=(None, '(*)', '\\1', '(*)', '\\2', '\\1'),
        valAttrNames=('jet_energy', 'muon_energy'),
        valIndices=('\\1', '\\2'),
    )

    event = MockEvent()
    event.ev = [ ]
    event.jet_pt = [ ]
    event.jet_eta = [ ]
    event.jet_phi = [ ]
    event.jet_energy = [ ]
    event.mu_pt = [ ]
    event.mu_eta = [ ]
    event.muon_energy = [ ]
    obj.begin(event)

    event.ev[:] = [1001]
    event.jet_pt[:]  =    [ 15.3, 12.9,  9.2, 10.5]
    event.jet_eta[:] =    [ -1.2,  5.2,  2.2,  0.5] # <- 2nd value is greater than max
    event.jet_phi[:] =    [  0.1,  0.6,  1.2] # <- the last value is missing
    event.jet_energy[:] = [ 16.2, 13.1, 10.1, 11.8]
    event.mu_pt[:]  =      [ 20.2, 11.9, 13.3,  5.2]
    event.mu_eta[:] =      [ 2.2,   1.2, -1.5, -0.5] # <- 1st value is greater than max
    event.muon_energy[:] = [ 22.1, 15.2, 16.3] # <- the last value is missing

    assert (
        ((1001, 15, -2, 11,  1, 0.1), (16.2, 15.2)),
        ((1001, 15, -2, 13, -2, 0.1), (16.2, 16.3)),
        ((1001,  9,  2, 11,  1, 1.2), (10.1, 15.2)),
        ((1001,  9,  2, 13, -2, 1.2), (10.1, 16.3)),
    ) == obj(event)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import math

from .Round import Round

##__________________________________________________________________||
class RoundLog(object):
    """Binning with equal width in log scale

    Parameters
    ----------
    width : float or int, default 1
        The common logarithm (log10) of the width.
    aboundary : float or int, optional
        A boundary. If not given, ``width/2`` will be used.
    min : float or int, optional
        The lowest bin will be the bin that ``min`` falls in. It must be a
        positive value. If given, ``__call__(val)`` returns ``underflow_bin``
        if the ``val`` is less than the lower edge of the lowest bin.
    underflow_bin : optional
        The underflow bin. When ``min`` is given, the ``__call__(val)`` returns
        ``underflow_bin`` if the ``val`` is less than the lower edge of the
        lowest bin.
    max : float or int, optional
        The highest bin will be the bin that ``max`` falls in except when
        ``max`` is one of boundaries. It must be a positive value. When ``max``
        is one of boundaries, the highest bin is the bin whose upper edge is
        ``max``. If given, ``__call__(val)`` returns the overflow bin if the
        ``val`` is greater than or equal to the upper edge of the highest bin.
    overflow_bin : optional
        The overflow bin if ``overflow_bin`` is any value other than ``True``.
        If ``overflow_bin`` is ``True``, the overflow bin will be the upper
        edge of the highest bin. When ``max`` is given, the ``__call__(val)``
        returns the overflow bin if the ``val`` is greater than or equal to the
        upper edge of the highest bin.
    valid : function, optional
        Boolean function to test if value is valid

    """
    def __init__(self, width=0.1, aboundary=1,
                 min=None, underflow_bin=None,
                 max=None, overflow_bin=None,
                 valid=None):

        self._round = Round(width=width, aboundary=math.log10(aboundary))
        self.width = width
        self.aboundary = aboundary
        self.min = min
        self.max = max
        self.valid = valid

        if self.min is None:
            self.min_bin_log10_lowedge = None
            self.underflow_bin = None
        else:
            self.min_bin_log10_lowedge = self._round(math.log10(self.min))
            self.underflow_bin = underflow_bin

        if self.max is None:
            self.max_bin_log10_upedge = None
            self.overflow_bin = None
        else:
            self._round(math.log10(self.max)) # = self._round.boundaries[-2]
            self.max_bin_log10_upedge = self._round.boundaries[-1]
            if overflow_bin is True:
                self.overflow_bin = 10**self.max_bin_log10_upedge
            else:
                self.overflow_bin = overflow_bin

    def __repr__(self):
        return '{}(width={!r}, aboundary={!r}, min={!r}, underflow_bin={!r}, max={!r}, overflow_bin={!r}, valid={!r})'.format(
            self.__class__.__name__,
            self.width,
            self.aboundary,
            self.min,
            self.underflow_bin,
            self.max,
            self.overflow_bin,
            self.valid
        )

    def __call__(self, val):

        if self.valid:
            if not self.valid(val):
                return None

        if val <= 0.0:
            if self.min is not None:
                return self.underflow_bin
            elif val == 0.0:
                return 0
            else:
                return None

        if self.min is not None:
            if math.log10(val) < self.min_bin_log10_lowedge:
                return self.underflow_bin

        if math.isinf(val):
            if self.max is not None:
                return self.overflow_bin
            else:
                return None

        if self.max is not None:
            if self.max_bin_log10_upedge <= math.log10(val):
                return self.overflow_bin

        val = math.log10(val)
        val = self._round(val)

        if val is None:
            return None

        return 10**val

    def next(self, bin):

        if bin is None:
            return None

        if bin == self.underflow_bin:
            return self.__call__(self.min)

        if bin < 0:
            return None

        if bin == 0:
            return 0

        if bin == self.overflow_bin:
            return self.overflow_bin

        log10_bin = self._round(math.log10(bin))

        if log10_bin is None:
            return None

        log10_next = self._round.next(log10_bin)

        if self.max is not None:
            if log10_next == self.max_bin_log10_upedge:
                return self.overflow_bin

        return 10**log10_next

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

import math
import collections
import logging

from .search import binary_search

##__________________________________________________________________||
class Round(object):
    """Equal width binning

    Parameters
    ----------
    width : float or int, default 1
        The width.
    aboundary : float or int, optional
        A boundary. If not given, ``width/2`` will be used.
    min : float or int, optional
        The lowest bin will be the bin that ``min`` falls in. If given,
        ``__call__(val)`` returns ``underflow_bin`` if the ``val`` is less than
        the lower edge of the lowest bin.
    underflow_bin : optional
        The underflow bin. When ``min`` is given, the ``__call__(val)`` returns
        ``underflow_bin`` if the ``val`` is less than the lower edge of the
        lowest bin.
    max : float or int, optional
        The highest bin will be the bin that ``max`` falls in except when
        ``max`` is one of boundaries. When ``max`` is one of boundaries, the
        highest bin is the bin whose upper edge is ``max``. If given,
        ``__call__(val)`` returns the overflow bin if the ``val`` is greater
        than or equal to the upper edge of the highest bin.
    overflow_bin : optional
        The overflow bin if ``overflow_bin`` is any value other than ``True``.
        If ``overflow_bin`` is ``True``, the overflow bin will be the upper
        edge of the highest bin. When ``max`` is given, the ``__call__(val)``
        returns the overflow bin if the ``val`` is greater than or equal to the
        upper edge of the highest bin.
    valid : function, optional
        Boolean function to test if value is valid

    """
    def __init__(self, width=1, aboundary=None,
                 min=None, underflow_bin=None,
                 max=None, overflow_bin=None,
                 valid=None):

        self.width = width
        self.aboundary = aboundary
        if aboundary is None:
            halfWidth = self.width//2 if self.width % 2 == 0 else float(self.width)/2
            aboundary = halfWidth
        self.boundaries = collections.deque([aboundary])
        self.min = min
        self.underflow_bin = underflow_bin
        self.max = max
        self.overflow_bin = overflow_bin
        self.valid = valid

        if self.min is not None:
            self._update_boundaries(self.min)

        if self.max is not None:
            self._update_boundaries(self.max)
            if self.overflow_bin is True:
                self.overflow_bin = self.boundaries[-1]

    def __repr__(self):
        return '{}(width={!r}, aboundary={!r}, min={!r}, underflow_bin={!r}, max={!r}, overflow_bin={!r}, valid={!r})'.format(
            self.__class__.__name__,
            self.width,
            self.aboundary,
            self.min,
            self.underflow_bin,
            self.max,
            self.overflow_bin,
            self.valid
        )

    def __call__(self, val):
        return self._lower_boundary(val)

    def _lower_boundary(self, val):

        if val is None:
            return None

        if self.valid:
            if not self.valid(val):
                return None

        if self.min is not None:
            if not self.boundaries[0] <= val:
                return self.underflow_bin

        if self.max is not None:
            if not val < self.boundaries[-1]:
                return self.overflow_bin

        if math.isinf(val):
            logger = logging.getLogger(__name__)
            logger.warning('val={}. will return {}'.format(val, None))
            return None

        self._update_boundaries(val)

        idx = binary_search(val, self.boundaries)
        return self.boundaries[idx]


    def _update_boundaries(self, val):

        while val < self.boundaries[0]:
            self.boundaries.appendleft(self.boundaries[0] - self.width)

        while val > self.boundaries[-1]:
            self.boundaries.append(self.boundaries[-1] + self.width)

    def next(self, bin):
        return self._next_lower_boundary(bin)

    def _next_lower_boundary(self, bin):

        bin = self._lower_boundary(bin)

        if bin is None:
            return None

        if bin == self.underflow_bin:
            return self._lower_boundary(self.min)

        if bin == self.overflow_bin:
            return self.overflow_bin

        self._update_boundaries(bin)

        return self._lower_boundary(bin + self.width*1.001)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

import alphatwirl

##__________________________________________________________________||
def test_logger_exist():
    assert 'alphatwirl' in logging.Logger.manager.loggerDict

def test_len_handlers():
    logger = logging.getLogger('alphatwirl')
    assert len(logger.handlers) >= 1

def test_logging():
    logger = logging.getLogger('alphatwirl')
    logger.error('test message')

##__________________________________________________________________||
def test_example():
    logger_names = logging.Logger.manager.loggerDict.keys()
    loglevel_dict = {l: logging.getLogger(l).getEffectiveLevel() for l in logger_names}

    # a dict of names and levels of loggers
    # e.g.,
    # {
    #     'alphatwirl': 40,
    #     'alphatwirl.delphes': 40,
    #     'alphatwirl.loop': 40,
    #     'pandas': 0,
    # }
    #
    # https://docs.python.org/3/library/logging.html#logging-levels
    #   Level Numeric value
    #   CRITICAL 50
    #   ERROR 40
    #   WARNING 30
    #   INFO 20
    #   DEBUG 10
    #   NOTSET 0

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import copy
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.loop import DatasetIntoEventBuildersSplitter

##__________________________________________________________________||
@pytest.fixture()
def MockEventBuilder():
    return mock.Mock()

@pytest.fixture()
def mockConfigMaker():
    ret = mock.Mock()
    file_list = ['A.root', 'B.root', 'C.root', 'D.root', 'E.root']
    def file_list_in(dataset, maxFiles):
        if maxFiles < 0:
            return file_list
        else:
            return file_list[:maxFiles]
    ret.file_list_in.side_effect = file_list_in
    ret.nevents_in_file.side_effect = [100, 200, 150, 180, 210]
    return ret

@pytest.fixture()
def obj(MockEventBuilder, mockConfigMaker):
    ret = DatasetIntoEventBuildersSplitter(
        EventBuilder=MockEventBuilder,
        eventBuilderConfigMaker=mockConfigMaker
    )
    return ret

def test_repr(obj):
   repr(obj)

def test_create_configs(obj, mockConfigMaker):
    dataset = mock.Mock()
    file_start_length_list = [('A.root', 0, 40), ('A.root', 40, 40), ('B.root', 0, 10)]
    config1 = mock.Mock()
    config2 = mock.Mock()
    config3 = mock.Mock()
    mockConfigMaker.create_config_for.side_effect = [config1, config2, config3]
    actual = obj._create_configs(dataset, file_start_length_list)
    assert [config1, config2, config3] == actual
    assert [
        mock.call(dataset, 'A.root', 0, 40),
        mock.call(dataset, 'A.root', 40, 40),
        mock.call(dataset, 'B.root', 0, 10),
    ] == mockConfigMaker.create_config_for.call_args_list

def test_call(obj):
    dataset = mock.Mock()
    obj(dataset)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.datasetloop import DatasetReaderComposite

##__________________________________________________________________||
@pytest.fixture()
def obj():
    return DatasetReaderComposite()

def test_repr(obj):
    repr(obj)

def test_one(obj):
    reader1 = mock.Mock()
    reader2 = mock.Mock()

    dataset1 = mock.Mock()
    dataset2 = mock.Mock()

    obj.add(reader1)
    obj.add(reader2)

    obj.begin()

    assert [mock.call()] == reader1.begin.call_args_list
    assert [mock.call()] == reader2.begin.call_args_list

    obj.read(dataset1)
    obj.read(dataset2)

    assert [mock.call(dataset1), mock.call(dataset2)] == reader1.read.call_args_list
    assert [mock.call(dataset1), mock.call(dataset2)] == reader2.read.call_args_list

    results = obj.end()
    assert [reader1.end(), reader2.end()] == results

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import logging
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree.inspect import get_entries_in_tree_in_file

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
def test_no_error():
    input_dirname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data_02')
    input_basename = 'sample_chain_02.root'
    path = os.path.join(input_dirname, input_basename)
    tree_name = 'tree'
    assert 1000 == get_entries_in_tree_in_file(path, tree_name)

def test_no_tree(caplog):
    input_dirname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data_02')
    input_basename = 'sample_chain_02.root'
    path = os.path.join(input_dirname, input_basename)
    tree_name = 'no_such_tree'

    with caplog.at_level(logging.WARNING):
        result = get_entries_in_tree_in_file(path, tree_name)

    assert result is None
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'inspect' in caplog.records[0].name
    assert 'cannot find' in caplog.records[0].msg

def test_no_tree_raise(caplog):
    input_dirname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data_02')
    input_basename = 'sample_chain_02.root'
    path = os.path.join(input_dirname, input_basename)
    tree_name = 'no_such_tree'

    with pytest.raises(ReferenceError):
        with caplog.at_level(logging.WARNING):
            get_entries_in_tree_in_file(path, tree_name, raises=True)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert 'inspect' in caplog.records[0].name
    assert 'cannot find' in caplog.records[0].msg

@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_zombie(caplog):
    input_dirname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data_02')
    input_basename = 'sample_chain_03_zombie.root'
    path = os.path.join(input_dirname, input_basename)
    tree_name = 'tree'

    with caplog.at_level(logging.WARNING):
        result = get_entries_in_tree_in_file(path, tree_name)

    assert result is None
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'inspect' in caplog.records[0].name
    assert 'cannot open' in caplog.records[0].msg

@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_zombie_raise(caplog):
    input_dirname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data_02')
    input_basename = 'sample_chain_03_zombie.root'
    path = os.path.join(input_dirname, input_basename)
    tree_name = 'tree'

    with pytest.raises(OSError):
        with caplog.at_level(logging.WARNING):
            get_entries_in_tree_in_file(path, tree_name, raises=True)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert 'inspect' in caplog.records[0].name
    assert 'cannot open' in caplog.records[0].msg

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection.factories.factory import LambdaStrFactory

##__________________________________________________________________||
def test_LambdaStrFactory():
    lambda_str = mock.Mock()
    MockClass = mock.Mock()
    components = ()
    name = mock.sentinel.name
    obj = LambdaStrFactory(lambda_str=lambda_str, LambdaStrClass=MockClass,
                           name=name, components=())
    assert [mock.call(lambda_str=lambda_str.format(), name=name)] == MockClass.call_args_list
    assert obj == MockClass()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from .modules.basic import All
from .modules.basic import Any
from .modules.basic import Not
from .modules.LambdaStr import LambdaStr
from .factories.factory import FactoryDispatcher

import os, sys

##__________________________________________________________________||
thisDir = os.path.dirname(os.path.realpath(__file__))
if not thisDir in sys.path: sys.path.append(thisDir)

##__________________________________________________________________||
def build_selection(**kargs):

    if 'aliasDict' not in kargs: kargs['aliasDict'] = { }

    if 'AllClass' not in kargs: kargs['AllClass'] = All
    if 'AnyClass' not in kargs: kargs['AnyClass'] = Any
    if 'NotClass' not in kargs: kargs['NotClass'] = Not
    if 'LambdaStrClass' not in kargs: kargs['LambdaStrClass'] = LambdaStr

    return FactoryDispatcher(**kargs)

##__________________________________________________________________||



from .BackrefMultipleArrayReader import BackrefMultipleArrayReader
from .Count import Count
from .KeyValueComposer import KeyValueComposer
from .NextKeyComposer import NextKeyComposer
from .Reader import Reader
from .Scan import Scan
from .Sum import Sum
from .Summarizer import Summarizer
from .WeightCalculatorOne import WeightCalculatorOne
from .parse_indices_config import parse_indices_config



# Tai Sakuma <tai.sakuma@gmail.com>
import copy
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.loop import ReaderComposite

##__________________________________________________________________||
@pytest.fixture()
def obj():
    return ReaderComposite()

def test_repr(obj):
    repr(obj)

def test_init_with_readers():
    reader1 = mock.Mock()
    reader2 = mock.Mock()
    readers = [reader1, reader2]
    obj = ReaderComposite(readers=readers)
    assert readers is not obj.readers
    assert readers == obj.readers

def test_event_two_readers_two_events(obj):
    """
    composite
        |- reader1
        |- reader2
    """
    reader1 = mock.Mock()
    reader2 = mock.Mock()
    obj.add(reader1)
    obj.add(reader2)

    events = mock.Mock()
    obj.begin(events)
    assert [mock.call(events)] == reader1.begin.call_args_list
    assert [mock.call(events)] == reader2.begin.call_args_list

    event1 = mock.Mock()
    obj.event(event1)

    event2 = mock.Mock()
    obj.event(event2)
    assert [mock.call(event1), mock.call(event2)], reader1.events.call_args_list
    assert [mock.call(event1), mock.call(event2)], reader2.events.call_args_list

    obj.end()
    assert [mock.call()] == reader1.end.call_args_list
    assert [mock.call()] == reader2.end.call_args_list

def test_event_nested_composite():
    """
    composite1
        |- composite2
        |      |- reader1
        |      |- reader2
        |- reader3
    """
    obj1 = ReaderComposite()
    obj2 = ReaderComposite()
    reader1 = mock.Mock()
    reader2 = mock.Mock()
    reader3 = mock.Mock()
    obj1.add(obj2)
    obj2.add(reader1)
    obj2.add(reader2)
    obj1.add(reader3)

    events = mock.Mock()
    obj1.begin(events)
    assert [mock.call(events)] == reader1.begin.call_args_list
    assert [mock.call(events)] == reader2.begin.call_args_list
    assert [mock.call(events)] == reader3.begin.call_args_list

    event1 = mock.Mock()
    obj1.event(event1)

    event2 = mock.Mock()
    obj1.event(event2)
    assert [mock.call(event1), mock.call(event2)], reader1.events.call_args_list
    assert [mock.call(event1), mock.call(event2)], reader2.events.call_args_list
    assert [mock.call(event1), mock.call(event2)], reader3.events.call_args_list

    obj1.end()
    assert [mock.call()] == reader1.end.call_args_list
    assert [mock.call()] == reader2.end.call_args_list
    assert [mock.call()] == reader3.end.call_args_list

def test_return_False(obj):
    """
    composite
        |- reader1 (return None)
        |- reader2 (return True)
        |- reader3 (return False)
        |- reader4
    """
    reader1 = mock.Mock()
    reader2 = mock.Mock()
    reader3 = mock.Mock()
    reader4 = mock.Mock()
    obj.add(reader1)
    obj.add(reader2)
    obj.add(reader3)
    obj.add(reader4)

    events = mock.Mock()
    obj.begin(events)

    reader1.event.return_value = None
    reader2.event.return_value = True
    reader3.event.return_value = False

    event1 = mock.Mock()
    ret = obj.event(event1)

    assert [mock.call(event1)], reader1.event.call_args_list
    assert [mock.call(event1)], reader2.event.call_args_list
    assert [mock.call(event1)], reader3.event.call_args_list
    assert [mock.call()], reader4.event.call_args_list
    assert ret is None

    obj.end()

def test_no_begin_end(obj):
    """
    composite
        |- reader1
        |- reader2 (without begin end)
        |- reader3
    """
    reader1 = mock.Mock()

    reader2 = mock.Mock()
    del reader2.begin
    del reader2.end

    reader3 = mock.Mock()
    obj.add(reader1)
    obj.add(reader2)
    obj.add(reader3)

    events = mock.Mock()
    obj.begin(events)
    assert [mock.call(events)] == reader1.begin.call_args_list
    assert [mock.call(events)] == reader3.begin.call_args_list

    event1 = mock.Mock()
    obj.event(event1)

    event2 = mock.Mock()
    obj.event(event2)
    assert [mock.call(event1), mock.call(event2)], reader1.events.call_args_list
    assert [mock.call(event1), mock.call(event2)], reader2.events.call_args_list
    assert [mock.call(event1), mock.call(event2)], reader3.events.call_args_list

    obj.end()
    assert [mock.call()] == reader1.end.call_args_list
    assert [mock.call()] == reader3.end.call_args_list

##__________________________________________________________________||
def test_merge(obj):
    """
    composite
        |- reader1
        |- reader2 (no merge)
        |- reader3
    """
    reader1 = mock.Mock()
    reader2 = mock.Mock()
    reader3 = mock.Mock()
    del reader2.merge

    obj.add(reader1)
    obj.add(reader2)
    obj.add(reader3)

    obj1 = copy.deepcopy(obj)
    assert obj.readers[0] is reader1
    assert obj.readers[1] is reader2
    assert obj.readers[2] is reader3
    assert obj1.readers[0] is not reader1
    assert obj1.readers[1] is not reader2
    assert obj1.readers[2] is not reader3

    obj.merge(obj1)

    assert [mock.call(obj1.readers[0])] == reader1.merge.call_args_list
    assert [mock.call(obj1.readers[2])] == reader3.merge.call_args_list

##__________________________________________________________________||
def test_collect(obj):
    """
    composite
        |- reader1
        |- reader2 (no collect)
        |- reader3
    """
    reader1 = mock.Mock()
    reader2 = mock.Mock()
    reader3 = mock.Mock()
    del reader2.collect

    obj.add(reader1)
    obj.add(reader2)
    obj.add(reader3)

    assert [reader1.collect(), None, reader3.collect()] == obj.collect()

##__________________________________________________________________||



from .Collector import Collector
from .CollectorComposite import CollectorComposite
from .CollectorDelegate import CollectorDelegate
from .DatasetIntoEventBuildersSplitter import DatasetIntoEventBuildersSplitter
from .EventLoop import EventLoop
from .EventLoopProgressReportWriter import EventLoopProgressReportWriter
from .EventLoopRunner import EventLoopRunner
from .EventDatasetReader import EventDatasetReader
from .MPEventLoopRunner import MPEventLoopRunner
from .NullCollector import NullCollector
from .ReaderComposite import ReaderComposite

from .EventsInDatasetReader import EventsInDatasetReader



import unittest
import inspect

import alphatwirl

##__________________________________________________________________||

class TestImport(unittest.TestCase):

    def test_functions(self):
        self.assertTrue(inspect.isfunction(alphatwirl.selection.build_selection))

    def test_classes(self):
        self.assertTrue(inspect.isclass(alphatwirl.selection.modules.All))
        self.assertTrue(inspect.isclass(alphatwirl.selection.modules.Any))
        self.assertTrue(inspect.isclass(alphatwirl.selection.modules.Not))
        self.assertTrue(inspect.isclass(alphatwirl.selection.modules.AllwCount))
        self.assertTrue(inspect.isclass(alphatwirl.selection.modules.AnywCount))
        self.assertTrue(inspect.isclass(alphatwirl.selection.modules.NotwCount))

##________________________________._________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import ROOT
import re

##__________________________________________________________________||
class BranchAddressManagerForVector(object):
    """The branch address manager for ROOT TTree

    This class manages ROOT.vector objects used for branch addresses
    of ROOT TTree. The main purpose of this class is to prevent
    multiple objects from being created for the same branch.

    All instances of this class share the data.
    """

    addressDict = { }

    def getVector(self, tree, branchName):
        """return the ROOT.vector object for the branch.

        """

        if (tree, branchName) in self.__class__.addressDict:
            return self.__class__.addressDict[(tree, branchName)]

        itsVector = self._getVector(tree, branchName)
        self.__class__.addressDict[(tree, branchName)] = itsVector

        return itsVector

    def _getVector(self, tree, branchName):

        leafNames = [l.GetName() for l in tree.GetListOfLeaves()]
        if branchName not in leafNames:
            return None

        leaf = tree.GetLeaf(branchName)
        typename = leaf.GetTypeName() # e.g., "vector<string>"
        match = re.search(r'^(.*)<(.*)>$', typename)
        if not match: return None
        if match.group(1) != 'vector': return None
        elementtypename = match.group(2) # e.g., "string", "int"

        tree.SetBranchStatus(leaf.GetName(), 1)
        itsVector = ROOT.vector(elementtypename)()
        tree.SetBranchAddress(leaf.GetName(), itsVector)

        return itsVector

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class MPEventLoopRunner(object):

    """This class (concurrently) runs instances of `EventLoop`.

    An instance of this class needs to be initialized with a
    communication channel with workers that actually run the
    `EventLoop`::

        runner = MPEventLoopRunner(communicationChannel)

    An example of a communication channel is an instance of
    `CommunicationChannel`.

    The method `begin()` does nothing in the current version::

        runner.begin()

    In older versions, multiple processes are forked in this method.

    Then, you can give an `EventLoop` with the method `run()`::

        runner.run(eventLoop1)

    This class will send the `EventLoop` to a worker through the
    communication channel. The worker, then, runs the `EventLoop`.

    You can call the method `run()` mutiple times::

        runner.run(eventLoop2)
        runner.run(eventLoop3)
        runner.run(eventLoop4)

    If workers are in the background, this method immediately returns.
    Worker are concurrently running the event loops in the background.
    If the worker is in the foreground, this method won't return until
    the worker finishes running the event loop. Whether workers are in
    the background or foreground depends on the communication channel
    with which this class is initialized.

    After giving all event loops that you need to run to this class,
    you need to call the method `end()`::

        results = runner.end()

    If workers are in the background, this method will wait until
    workers finish running all event loops. If the worker is in the
    foreground, this method immediately returns. This method returns
    the results, the list of the values eventLoops return, sorted in
    the order given with `run()`.

    """

    def __init__(self, communicationChannel):
        self.communicationChannel = communicationChannel
        self.nruns = 0

    def __repr__(self):
        return '{}(communicationChannel = {!r}'.format(
            self.__class__.__name__,
            self.communicationChannel
        )

    def begin(self):
        """does nothing.

        Older versions of this class had implementations.
        """
        pass

    def run(self, eventLoop):
        """run the event loop in the background.

        Args:
            eventLoop (EventLoop): an event loop to run

        """

        self.nruns += 1
        return self.communicationChannel.put(eventLoop)

    def run_multiple(self, eventLoops):
        """run the event loops in the background.

        Args:
            eventLoops (list): a list of event loops to run

        """

        self.nruns += len(eventLoops)
        return self.communicationChannel.put_multiple(eventLoops)

    def poll(self):
        """Return pairs of run ids and results of finish event loops.
        """
        ret = self.communicationChannel.receive_finished()
        self.nruns -= len(ret)
        return ret

    def receive_one(self):
        """Return a pair of a run id and a result.

        This method waits until an event loop finishes.
        This method returns None if no loop is running.
        """
        if self.nruns == 0:
            return None
        ret = self.communicationChannel.receive_one()
        if ret is not None:
            self.nruns -= 1
        return ret

    def receive(self):
        """Return pairs of run ids and results.

        This method waits until all event loops finish
        """
        ret = self.communicationChannel.receive_all()
        self.nruns -= len(ret)
        if self.nruns > 0:
            import logging
            logger = logging.getLogger(__name__)
            logger.warning(
                'too few results received: {} results received, {} more expected'.format(
                    len(ret), self.nruns))
        elif self.nruns < 0:
            import logging
            logger = logging.getLogger(__name__)
            logger.warning(
                'too many results received: {} results received, {} too many'.format(
                    len(ret), -self.nruns))
        return ret

    def end(self):
        """wait until all event loops end and returns the results.

        """

        results = self.communicationChannel.receive()

        if self.nruns != len(results):
            import logging
            logger = logging.getLogger(__name__)
            # logger.setLevel(logging.DEBUG)
            logger.warning(
                'too few results received: {} results received, {} expected'.format(
                    len(results),
                    self.nruns
                ))

        return results

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time
import sys, collections

##__________________________________________________________________||
class ProgressPrint(object):
    def __init__(self):
        self.reports = collections.OrderedDict()
        self.lines = [ ]
        self.interval = 60.0 # [second]
        self._readTime()
        self.last = [ ]

    def __repr__(self):
        return '{}()'.format(
            self.__class__.__name__
        )

    def nreports(self):
        return len(self.reports)

    def present(self, report):
        self.reports[report.taskid] = report
        if not self._need_to_update(report): return
        self._create_lines()
        self._print_lines()
        self._readTime()

    def _create_lines(self):
        self.lines = [ ]
        taskids_to_delete = [ ]
        for taskid, report in self.reports.items():
            line = self.createLine(report)
            if report.done >= report.total:
                taskids_to_delete.append(report.taskid)
                self.last.append(line)
            else:
                self.lines.append(line)
        for taskid in taskids_to_delete:
            del self.reports[taskid]

    def _print_lines(self):
        sys.stdout.write("\n")
        sys.stdout.write(time.asctime(time.localtime(time.time())))
        sys.stdout.write("\n")
        if len(self.last) > 0: sys.stdout.write("\n".join(self.last) + "\n")
        sys.stdout.write("\n".join(self.lines) + "\n")
        sys.stdout.flush()

    def createLine(self, report):
        percent = float(report.done)/report.total if report.total > 0 else 1
        bar = (':' * int(percent * 40)).ljust(40, " ")
        percent = round(percent * 100, 2)
        return " {1:8d} / {2:8d} ({0:6.2f}%) {3} ".format(percent, report.done, report.total, report.name)

    def _need_to_update(self, report):
        if self._time() - self.lastTime > self.interval: return True
        if report.done == report.total: return True
        if report.done == 0: return True
        return False

    def _time(self): return time.time()
    def _readTime(self): self.lastTime = self._time()

##__________________________________________________________________||



#!/usr/bin/env python
# Tai Sakuma <tai.sakuma@gmail.com>
import os, sys
import argparse
import tarfile
import gzip

try:
    import cPickle as pickle
except:
    import pickle

##__________________________________________________________________||
parser = argparse.ArgumentParser()
parser.add_argument('path', nargs=1, help='path to pickle')
args = parser.parse_args()

##__________________________________________________________________||
def main():

    cwd = os.getcwd()
    workingarea_path = os.path.dirname(args.path[0])
    pickle_basename = os.path.basename(args.path[0])

    setup(cwd, workingarea_path)

    run(cwd, workingarea_path, pickle_basename)

##__________________________________________________________________||
def run(cwd, workingarea_path, pickle_basename):

    os.chdir(workingarea_path)

    with gzip.open(pickle_basename, 'rb') as f:
        reader = pickle.load(f)

    os.chdir(cwd)
    reader.end()

##__________________________________________________________________||
def setup(cwd, workingarea_path):

    os.chdir(workingarea_path)

    dirname = 'python_modules'
    tarname = dirname + '.tar.gz'

    if os.path.exists(tarname) and not os.path.exists(dirname):
        if try_make_file('.untarring'):
            tar = tarfile.open(tarname)
            tar.extractall()
            tar.close()
            os.remove('.untarring')

    while os.path.isfile('.untarring'):
       pass

    if os.path.exists(dirname):
        sys.path.insert(0, dirname)

    os.chdir(cwd)

##__________________________________________________________________||
# http://stackoverflow.com/questions/33223564/atomically-creating-a-file-if-it-doesnt-exist-in-python
def try_make_file(filename):
    try:
        os.open(filename,  os.O_CREAT | os.O_EXCL)
        return True
    except OSError:
        # FileExistsError can be used for Python 3
        return False

##__________________________________________________________________||
if __name__ == '__main__':
    main()



# Tai Sakuma <tai.sakuma@gmail.com>
import time

##__________________________________________________________________||
class ProgressReporter(object):
    """A progress reporter

    This class sends a `ProgressReport` to a progress monitor, e.g.,
    `ProgressMonitor` or `BProgressMonitor`, which, for example, uses
    the reports to update `ProgressBar` on the screen.

    An instance of this class is initialized with a message queue::

        reporter = ProgressReporter(queue)

    A reporter, an instance of this class, is typically created and
    initialized by a progress monitor (`ProgressMonitor` or
    `BProgressMonitor`), which keeps the other end of the message
    queue. A reporter and a monitor might be running in different
    processes.

    A report, an instance of `ProgressReport`, can be sent as::

        reporter.report(report)

    This method can be frequently called multiple times. However,
    after sending one report, the reporter wait for a certain
    ``interval`` (0.1 seconds by default) before sending another
    report. Reports received within this interval will be discarded.
    The exception for this is the last report. The last report, which
    indicates the completion of the task, will be always sent to the
    progress monitor regardless of whether it is given within the
    interval.

    """
    def __init__(self, queue):
        self.queue = queue
        self.interval = 0.1 # [second]
        self._read_time()

    def __repr__(self):
        return '{}(queue={!r}, interval={!r})'.format(
            self.__class__.__name__,
            self.queue,
            self.interval
        )

    def report(self, report):
        """send ``report`` to a progress monitor

        Args:
            report (ProgressReport): a progress report

        """

        if not self._need_to_report(report): return
        self._report(report)

    def _report(self, report):
        self.queue.put(report)
        self._read_time()

    def _need_to_report(self, report):
        if self._time() - self.last_time > self.interval: return True
        if report.last(): return True
        if report.first(): return True
        return False

    def _read_time(self): self.last_time = self._time()

    def _time(self): return time.time()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

from .WeightCalculatorOne import WeightCalculatorOne

##__________________________________________________________________||
class Reader(object):
    def __init__(self, keyValComposer, summarizer,
                 nextKeyComposer=None,
                 weightCalculator=WeightCalculatorOne(),
                 collector=None, nevents=None):

        self.keyValComposer = keyValComposer
        self.summarizer = summarizer
        self.collector = collector
        self.weightCalculator = weightCalculator
        self.nextKeyComposer = nextKeyComposer

        self.nevents = nevents
        self.ievent = 0

        self._repr_pairs = [
            ('keyValComposer', self.keyValComposer),
            ('summarizer', self.summarizer),
            ('collector', self.collector),
            ('nextKeyComposer', self.nextKeyComposer),
            ('weightCalculator', self.weightCalculator),
            ('nevents', self.nevents),
        ]

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in self._repr_pairs]),
        )

    def __str__(self):
        nwidth = max(len(n) for n, _ in self._repr_pairs)
        nwidth += 4
        return '{}:\n{}'.format(
            self.__class__.__name__,
            '\n'.join(['{:>{}}: {!r}'.format(n, nwidth, v) for n, v in self._repr_pairs]),
        )

    def begin(self, event):
        self.keyValComposer.begin(event)

    def event(self, event):
        if self.nevents is not None and self.nevents <= self.ievent:
            return

        self.ievent += 1

        try:
            keyvals = self.keyValComposer(event)
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.error(e)
            logger.error(self)
            raise

        weight = self.weightCalculator(event)
        for key, val in keyvals:
            self.summarizer.add(key=key, val=val, weight=weight)

    def end(self):
        if self.nextKeyComposer is None:
            return

        for key in sorted(self.summarizer.keys()):
            nextKeys = self.nextKeyComposer(key)
            for nextKey in nextKeys:
                self.summarizer.add_key(nextKey)

    def merge(self, other):
        self.summarizer += other.summarizer

    def results(self):
        return self.summarizer

    def collect(self):
        return self.collector(self)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
import collections

##__________________________________________________________________||
# EventBuilderConfig is deprecated as EventBuilder is being replaced
# with BuildEvents, which uses dict instead of EventBuilderConfig

EventBuilderConfig = collections.namedtuple(
    'EventBuilderConfig',
    'inputPaths treeName maxEvents start name'
)
# 'name' is used by loop.EventLoopProgressReportWriter

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from ..misc import mkdir_p
import os

##__________________________________________________________________||
class WritePandasDataFrameToFile(object):
    def __init__(self, outPath):
        self._outPath = outPath

    def deliver(self, results):
        if results is None: return
        f = self._open(self._outPath)
        if len(results.index) == 0:
            towrite = " ".join([i for i in results.columns]) + "\n"
            towrite = towrite.encode()
            f.write(towrite)
        else:
            ## results.to_string(f, index = False)
            towrite = results.to_string(index = False) + "\n"
            towrite = towrite.encode()
            f.write(towrite)
        self._close(f)

    def _open(self, path):
        mkdir_p(os.path.dirname(path))
        return open(path, 'w')

    def _close(self, file): file.close()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import pytest

##__________________________________________________________________||
# @pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")
@pytest.fixture(autouse=True)
def monkeypatch_class_attributes_BranchBuilder(monkeypatch):
    ## monkey patch a class attribute of BranchBuilder.
    ## segmentation violation can happen without this patch.
    ## TODO probably need similar patches for
    ## any other classes with class attributes.
    try:
        module = sys.modules['alphatwirl.roottree.BranchBuilder']
    except KeyError:
        # the exception can happen if ROOT cannot be imported
        yield
        return

    ret =  { }
    monkeypatch.setattr(module.BranchBuilder, 'itsdict', ret)
    yield ret

##__________________________________________________________________||




##__________________________________________________________________||
try:
    ## https://root-forum.cern.ch/t/pyroot-hijacks-help/15207
    import ROOT
    ROOT.PyConfig.IgnoreCommandLineOptions = True
except ImportError:
    pass

##__________________________________________________________________||

from . import binning
from . import collector
from . import concurrently
from . import configure
from . import roottree
from . import selection
from . import datasetloop
from . import loop
from . import parallel
from . import progressbar
from . import summary
from .misc import mkdir_p
from .misc import list_to_aligned_text
from .misc import quote_string

# to be deleted
from .misc import listToAlignedText

##__________________________________________________________________||
import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.ERROR) # DEBUG INFO WARN ERROR CRITICAL
if not logger.handlers:
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    del handler
    del formatter
del logger

##__________________________________________________________________||

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.loop.splitfuncs import create_files_start_length_list
from alphatwirl.loop.splitfuncs import _apply_max_files
from alphatwirl.loop.splitfuncs import _file_nevents_list
from alphatwirl.loop.splitfuncs import _fast_path

##__________________________________________________________________||
@pytest.fixture()
def wrapped_apply_max_files(monkeypatch):
    module = sys.modules['alphatwirl.loop.splitfuncs']
    ret = mock.Mock(wraps=_apply_max_files)
    monkeypatch.setattr(module, '_apply_max_files', ret)
    return ret

@pytest.fixture()
def mock_fast_path(monkeypatch):
    module = sys.modules['alphatwirl.loop.splitfuncs']
    ret = mock.Mock(name='mock_fast_path')
    monkeypatch.setattr(module, '_fast_path', ret)
    return ret

@pytest.fixture()
def mock_full_path(monkeypatch):
    module = sys.modules['alphatwirl.loop.splitfuncs']
    ret = mock.Mock(name='mock_full_path')
    monkeypatch.setattr(module, '_full_path', ret)
    return ret

@pytest.mark.parametrize('files', [[], ['A.root'], ['A.root', 'B.root', 'C.root']])
@pytest.mark.parametrize('max_events', [-1, 0, 1, 100])
@pytest.mark.parametrize('max_events_per_run', [-1, 0, 1, 100])
@pytest.mark.parametrize('max_files', [-1, 0, 1, 100])
@pytest.mark.parametrize('max_files_per_run', [-1, 0, 1, 100])
def test_create_files_start_length_list(
        files, max_events, max_events_per_run,
        max_files, max_files_per_run,
        wrapped_apply_max_files,
        mock_fast_path, mock_full_path
):
    actual = create_files_start_length_list(
        files,
        func_get_nevents_in_file=None,
        max_events=max_events,
        max_events_per_run=max_events_per_run,
        max_files=max_files,
        max_files_per_run=max_files_per_run
    )

    wrapped_apply_max_files.assert_called_once()

    if max_events == 0 or max_events_per_run == 0:
        assert [ ] == actual
    elif max_events < 0 and max_events_per_run < 0:
        assert mock_fast_path() == actual
    else:
        assert mock_full_path() == actual

##__________________________________________________________________||
def test_create_files_start_length_list_default():
    file1 = mock.sentinel.file1
    file2 = mock.sentinel.file2
    file3 = mock.sentinel.file3
    files = [file1, file2, file3]
    actual = create_files_start_length_list(files)
    assert [([file1], 0, -1), ([file2], 0, -1), ([file3], 0, -1)] == actual

##__________________________________________________________________||
@pytest.mark.parametrize('args, expected', [

    ## empty
    pytest.param(([ ], -1), [ ], id='empty. -1'),
    pytest.param(([ ],  0), [ ], id='empty. 0'),
    pytest.param(([ ],  1), [ ], id='empty. 1'),
    pytest.param(([ ], 10), [ ], id='empty. 10'),
    pytest.param((['A.root' ], -1), ['A.root'], id='1 file. -1'),
    pytest.param((['A.root' ],  0), [ ], id='1 file. 0'),
    pytest.param((['A.root' ],  1), ['A.root'], id='1 file. 1'),
    pytest.param((['A.root' ], 10), ['A.root'], id='1 file. 10'),
    pytest.param(
        (
            ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
            -1
        ),
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        id='all'
    ),
    pytest.param(
        (['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], 0),
        [ ],
        id='zero'
    ),
    pytest.param(
        (['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], 1),
        ['A.root'],
        id='one'
    ),
    pytest.param(
        (['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], 3),
        ['A.root', 'B.root', 'C.root'],
        id='smaller max_events'
    ),
    pytest.param(
        (['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], 5),
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        id='exact max_events'
    ),
    pytest.param(
        (['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], 10),
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        id='larger max_events'
    ),
])
def test_apply_max_files(args, expected):
    assert expected == _apply_max_files(*args)

##__________________________________________________________________||
@pytest.mark.parametrize(
    'files, nevents, max_events, expected_results, expected_call_args', [
    pytest.param([ ], [ ], -1, [ ], [ ], id='empty-no-max'),
    pytest.param([ ], [ ], 10, [ ], [ ], id='empty-with-max'),
    pytest.param(
        ['A.root'], [100], -1, [('A.root', 100)],
        ['A.root'],
        id='one-file-no-max'
    ),
    pytest.param(
        ['A.root'], [100], 0, [ ],
        [ ],
        id='one-file-zero-max'
    ),
    pytest.param(
        ['A.root'], [0], -1, [ ],
        ['A.root'],
        id='one-file-empty'
    ),
    pytest.param(
        ['A.root'], [None], -1, [ ],
        ['A.root'],
        id='one-file-error'
    ),
    pytest.param(
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        [100, 200, 150, 180, 210],
        -1,
        [('A.root', 100), ('B.root', 200), ('C.root', 150), ('D.root', 180), ('E.root', 210)],
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        id='five-files-no-max'
    ),
    pytest.param(
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        [100, 200, 150, 180, 210],
        0,
        [ ],
        [ ],
        id='five-files-zero-max'
    ),
    pytest.param(
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        [100, 200, 150, 180, 210],
        150,
        [('A.root', 100), ('B.root', 200)],
        ['A.root', 'B.root'],
        id='five-files-with-max'
    ),
    pytest.param(
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        [100, 200, 150, 180, 210],
        300, # exactly the nevents in the first two files
        [('A.root', 100), ('B.root', 200)],
        ['A.root', 'B.root'],
        id='five-files-max-exactly-first-two-files'
    ),
    pytest.param(
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        [100, 0, 150, 180, 210],
        -1,
        [('A.root', 100), ('C.root', 150), ('D.root', 180), ('E.root', 210)],
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        id='five-files-one-empty-file'
    ),
    pytest.param(
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        [0, 0, 0, 0, 0],
        -1,
        [ ],
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        id='five-files-all-empty'
    ),
    pytest.param(
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        [100, None, 150, 180, 210],
        -1,
        [('A.root', 100), ('C.root', 150), ('D.root', 180), ('E.root', 210)],
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        id='five-files-one-error-file'
    ),
    pytest.param(
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        [None, None, None, None, None],
        -1,
        [ ],
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'],
        id='five-files-all-error'
    ),
])
def test_file_nevents_list_(
        files, nevents, max_events,
        expected_results, expected_call_args):
    func_get_nevents_in_file = mock.Mock()
    func_get_nevents_in_file.side_effect = nevents
    actual = _file_nevents_list(
        files,
        func_get_nevents_in_file=func_get_nevents_in_file,
        max_events=max_events
    )
    assert expected_results == actual
    assert [mock.call(a) for a in expected_call_args] == func_get_nevents_in_file.call_args_list

##__________________________________________________________________||
@pytest.mark.parametrize('files, max_files_per_run, expected', [
    (
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], -1,
        [(['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], 0, -1)]
    ),
    (
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], 0,
        [ ]
    ),
    (
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], 1,
        [
            (['A.root'], 0, -1),
            (['B.root'], 0, -1),
            (['C.root'], 0, -1),
            (['D.root'], 0, -1),
            (['E.root'], 0, -1)
        ]
    ),
    (
        ['A.root', 'B.root', 'C.root', 'D.root', 'E.root'], 2,
        [
            (['A.root', 'B.root'], 0, -1),
            (['C.root', 'D.root'], 0, -1),
            (['E.root'], 0, -1)
        ]
    ),
])
def test_fast_path(files, max_files_per_run, expected):
    actual = _fast_path(files, max_files_per_run)
    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import logging

import pytest

try:
   import cPickle as pickle
except:
   import pickle

from alphatwirl.misc.removal import _removed

##__________________________________________________________________||
@_removed(msg='extra message')
def func():
    pass

def test_func_logging(caplog):
    with pytest.raises(RuntimeError):
       with caplog.at_level(logging.ERROR):
          func()

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert 'test_removal' in caplog.records[0].name
    assert 'func() is removed. extra message' in caplog.records[0].msg

def test_func_name():
    assert  'func' == func.__name__

def test_func_pickle():
    pickle.dumps(func)

@_removed()
def func_without_msg():
    pass

def test_func_logging_without_msg(caplog):
    with pytest.raises(RuntimeError):
       with caplog.at_level(logging.ERROR):
          func_without_msg()

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert 'test_removal' in caplog.records[0].name
    assert 'func_without_msg() is removed.' in caplog.records[0].msg

##__________________________________________________________________||
@_removed(msg='extra message')
class ClassWithInit(object):
    def __init__(self):
        pass

@_removed(msg='extra message')
class ClassWithoutInit(object):
    pass

@_removed()
class ClassWithInitNoMsg(object):
    def __init__(self):
        pass

@_removed()
class ClassWithoutInitNoMsg(object):
    pass

@pytest.mark.parametrize('Class', (ClassWithInit, ClassWithoutInit))
def test_class_logging(Class, caplog):
    with pytest.raises(RuntimeError):
       with caplog.at_level(logging.ERROR):
          c = Class()
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert 'test_removal' in caplog.records[0].name
    expected = '{} is removed. extra message'.format(Class.__name__)
    assert expected in caplog.records[0].msg

@pytest.mark.parametrize('Class', (ClassWithInitNoMsg, ClassWithoutInitNoMsg))
def test_class_logging_no_msg(Class, caplog):
    with pytest.raises(RuntimeError):
       with caplog.at_level(logging.ERROR):
          c = Class()
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert 'test_removal' in caplog.records[0].name
    expected = '{} is removed.'.format(Class.__name__)
    assert expected in caplog.records[0].msg

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection.modules import AllwCount, AnywCount, NotwCount

##__________________________________________________________________||
@pytest.mark.parametrize('Class', [AllwCount, AnywCount],
                         ids=[AllwCount.__name__, AnywCount.__name__])
def test_allany_init_collector(Class):
    obj = Class()
    assert obj.collector is None
    col = mock.sentinel.col
    obj = Class(collector=col)
    assert obj.collector is col

def test_not_init_collector():
    sel = mock.Mock()
    obj = NotwCount(selection=sel)
    assert obj.collector is None
    col = mock.sentinel.col
    obj = NotwCount(selection=sel, collector=col)
    assert obj.collector is col

##__________________________________________________________________||
class MockEventSelection(object):
    def begin(self, event): pass
    def __call__(self, event): pass
    def end(self): pass

##__________________________________________________________________||
@pytest.mark.parametrize('method', [1, 2, 3])
def test_all(method):
    sel1 = mock.Mock(spec=MockEventSelection)
    sel2 = mock.Mock(spec=MockEventSelection)
    sel1.name ='sel1'
    sel2.name =''

    if method == 1:
        obj = AllwCount()
        obj.add(sel1)
        obj.add(sel2)
    elif method == 2:
        obj = AllwCount(selections=[sel1, sel2])
    elif method == 3:
        obj = AllwCount(selections=[sel1])
        obj.add(sel2)

    event = mock.Mock()

    obj.begin(event)

    # need to set return_value each time because a selection is not
    # guaranteed to be called

    sel1.return_value = True
    sel2.return_value = True
    assert obj(event)

    sel1.return_value = True
    sel2.return_value = False
    assert not obj(event)

    sel1.return_value = False
    sel2.return_value = True
    assert not obj(event)

    sel1.return_value = False
    sel2.return_value = False
    assert not obj(event)

    sel1.return_value = False
    sel2.return_value = False
    assert not obj(event)

    obj.end()

    count = obj.results()
    assert [
            [1, 'MockEventSelection', 'sel1', 2, 5],
            [1, 'MockEventSelection',     '', 1, 2],
    ] == count._results

##__________________________________________________________________||
@pytest.mark.parametrize('method', [1, 2, 3])
def test_any(method):
    sel1 = mock.Mock(spec=MockEventSelection)
    sel2 = mock.Mock(spec=MockEventSelection)
    sel1.name ='sel1'
    sel2.name =''

    if method == 1:
        obj = AnywCount()
        obj.add(sel1)
        obj.add(sel2)
    elif method == 2:
        obj = AnywCount(selections=[sel1, sel2])
    elif method == 3:
        obj = AnywCount(selections=[sel1])
        obj.add(sel2)

    event = mock.Mock()

    obj.begin(event)

    # need to set return_value each time because a selection is not
    # guaranteed to be called

    sel1.return_value = True
    sel2.return_value = True
    assert obj(event)

    sel1.return_value = True
    sel2.return_value = False
    assert obj(event)

    sel1.return_value = False
    sel2.return_value = True
    assert obj(event)

    sel1.return_value = False
    sel2.return_value = False
    assert not obj(event)

    sel1.return_value = False
    sel2.return_value = False
    assert not obj(event)


    obj.end()

    count = obj.results()
    assert [
            [1, 'MockEventSelection', 'sel1', 2, 5],
            [1, 'MockEventSelection',     '', 1, 3],
    ] == count._results

##__________________________________________________________________||
def test_not():
    sel1 = mock.Mock(spec=MockEventSelection)
    sel1.name ='sel1'

    obj = NotwCount(sel1)

    event = mock.Mock()

    obj.begin(event)

    sel1.return_value = True
    assert not obj(event)

    sel1.return_value = False
    assert obj(event)

    sel1.return_value = False
    assert obj(event)

    obj.end()

    count = obj.results()
    assert [
        [1, 'MockEventSelection', 'sel1', 1, 3],
    ] == count._results

##__________________________________________________________________||
def test_all_empty():
    obj = AllwCount()
    event = mock.Mock()
    obj.begin(event)
    assert obj(event)
    obj.end()
    count = obj.results()
    assert [ ] == count._results

def test_any_empty():
    obj = AnywCount()
    event = mock.Mock()
    obj.begin(event)
    assert not obj(event)
    obj.end()
    count = obj.results()
    assert [ ] == count._results

##__________________________________________________________________||
@pytest.mark.parametrize(
    'Class', [AllwCount, AnywCount],
    ids=[AllwCount.__name__, AnywCount.__name__])
def test_allany_collector(Class):
    obj = Class()
    obj.collect()

    collector = mock.Mock()
    obj = Class(collector=collector)
    result = obj.collect()
    assert [mock.call(obj)] == collector.call_args_list
    assert result is collector()

def test_not_collector():
    sel = mock.Mock()
    obj = NotwCount(selection=sel)
    obj.collect()

    collector = mock.Mock()
    obj = NotwCount(selection=sel, collector=collector)
    result = obj.collect()
    assert [mock.call(obj)] == collector.call_args_list
    assert result is collector()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import time
import collections

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.concurrently import CommunicationChannel
from alphatwirl.concurrently import CommunicationChannel0
from alphatwirl.concurrently import MultiprocessingDropbox
from alphatwirl.concurrently import WorkingArea
from alphatwirl.concurrently import SubprocessRunner
from alphatwirl.concurrently import TaskPackageDropbox

from alphatwirl.concurrently.testing import MockResult, MockTask

##__________________________________________________________________||
@pytest.fixture(
    params=[
        'CommunicationChannel0',
        'CommunicationChannel-multiprocessing',
        ## 'CommunicationChannel-subprocess'
    ])
def obj(request, tmpdir_factory, monkeypatch):
    name = request.param
    if name == 'CommunicationChannel0':
        ret = CommunicationChannel0()
    elif name == 'CommunicationChannel-multiprocessing':
        dropbox = MultiprocessingDropbox()
        ret = CommunicationChannel(dropbox=dropbox)
    elif name == 'CommunicationChannel-subprocess':
        topdir = str(tmpdir_factory.mktemp(''))
        topdir = os.path.join(topdir, '_ccsp_temp')
        workingarea = WorkingArea(topdir=topdir, python_modules=( ))
        dropbox = TaskPackageDropbox(
            workingArea=workingarea,
            dispatcher=SubprocessRunner()
        )
        ret = CommunicationChannel(dropbox=dropbox)

    ret.begin()
    yield ret
    ret.end()

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_extra_begin(obj):
    obj.begin()

def test_extra_end(obj):
    obj.end()

def test_terminate(obj):
    obj.terminate()

##__________________________________________________________________||
task1 = MockTask(name='task1', time=0.01)
task1_args = (123, 'ABC')
task1_kwargs = {'A': 34}
result1 = MockResult(name='task1', args=task1_args, kwargs=task1_kwargs)

task2 = MockTask(name='task2', time=0.001)
task2_args = ( )
task2_kwargs = { }
result2 = MockResult(name='task2', args=task2_args, kwargs=task2_kwargs)

task3 = MockTask(name='task3', time=0.005)
task3_args = (33, 44)
task3_kwargs = { }
result3 = MockResult(name='task3', args=task3_args, kwargs=task3_kwargs)

task4 = MockTask(name='task4', time=0.002)
task4_args = ( )
task4_kwargs = dict(ABC='abc', DEF='def')
result4 = MockResult(name='task4', args=task4_args, kwargs=task4_kwargs)

##__________________________________________________________________||
def test_put_receive_one_task(obj):
    assert 0 == obj.put(task1, *task1_args, **task1_kwargs)
    assert [result1] == obj.receive()

def test_receive_without_put(obj):
    assert [ ] == obj.receive()

def test_put_receive_four_tasks(obj):
    assert 0 == obj.put(task1, *task1_args, **task1_kwargs)
    assert 1 == obj.put(task2, *task2_args, **task2_kwargs)
    assert 2 == obj.put(task3, *task3_args, **task3_kwargs)
    assert 3 == obj.put(task4, *task4_args, **task4_kwargs)

    assert [result1, result2, result3, result4] == obj.receive()
    # sorted in the order put

def test_put_receive_repeat(obj):
    assert 0 == obj.put(task1, *task1_args, **task1_kwargs)
    assert 1 == obj.put(task2, *task2_args, **task2_kwargs)
    assert [result1, result2] == obj.receive()
    assert 2 == obj.put(task3, *task3_args, **task3_kwargs)
    assert 3 == obj.put(task4, *task4_args, **task4_kwargs)
    assert [result3, result4] == obj.receive()

def test_put_receive_end_repeat(obj):
    assert 0 == obj.put(task1, *task1_args, **task1_kwargs)
    assert 1 == obj.put(task2, *task2_args, **task2_kwargs)
    assert [result1, result2] == obj.receive()
    obj.end()
    obj.begin()
    assert 2 == obj.put(task3, *task3_args, **task3_kwargs)
    assert 3 == obj.put(task4, *task4_args, **task4_kwargs)
    assert [result3, result4] == obj.receive()

##__________________________________________________________________||
def test_put_multiple_receive_empty(obj):
    assert [ ] == obj.put_multiple([ ])
    assert [ ] == obj.receive()

def test_put_multiple_receive_four_tasks(obj):
    assert [0, 1, 2, 3] == obj.put_multiple([
        dict(task=task1, args=task1_args, kwargs=task1_kwargs),
        dict(task=task2, args=task2_args, kwargs=task2_kwargs),
        dict(task=task3, args=task3_args, kwargs=task3_kwargs),
        dict(task=task4, args=task4_args, kwargs=task4_kwargs),
    ])

    assert [result1, result2, result3, result4] == obj.receive()
    # sorted in the order put

def test_put_multiple_receive_repeat(obj):
    assert [0, 1] == obj.put_multiple([
        dict(task=task1, args=task1_args, kwargs=task1_kwargs),
        dict(task=task2, args=task2_args, kwargs=task2_kwargs),
    ])
    assert [result1, result2] == obj.receive()
    assert [2, 3] == obj.put_multiple([
        dict(task=task3, args=task3_args, kwargs=task3_kwargs),
        dict(task=task4, args=task4_args, kwargs=task4_kwargs),
    ])
    assert [result3, result4] == obj.receive()

def test_put_multiple_receive_end_repeat(obj):
    assert [0, 1] == obj.put_multiple([
        dict(task=task1, args=task1_args, kwargs=task1_kwargs),
        dict(task=task2, args=task2_args, kwargs=task2_kwargs),
    ])
    assert [result1, result2] == obj.receive()
    obj.end()
    obj.begin()
    assert [2, 3] == obj.put_multiple([
        dict(task=task3, args=task3_args, kwargs=task3_kwargs),
        dict(task=task4, args=task4_args, kwargs=task4_kwargs),
    ])
    assert [result3, result4] == obj.receive()

##__________________________________________________________________||
def test_put_multiple_receive_all_empty(obj):
    assert [ ] == obj.put_multiple([ ])
    assert [ ] == obj.receive_all()

def test_put_multiple_receive_all_four_tasks(obj):
    assert [0, 1, 2, 3] == obj.put_multiple([
        dict(task=task1, args=task1_args, kwargs=task1_kwargs),
        dict(task=task2, args=task2_args, kwargs=task2_kwargs),
        dict(task=task3, args=task3_args, kwargs=task3_kwargs),
        dict(task=task4, args=task4_args, kwargs=task4_kwargs),
    ])

    assert [(0, result1), (1, result2), (2, result3), (3, result4)] == obj.receive_all()
    # sorted in the order put

def test_put_multiple_receive_all_repeat(obj):
    assert [0, 1] == obj.put_multiple([
        dict(task=task1, args=task1_args, kwargs=task1_kwargs),
        dict(task=task2, args=task2_args, kwargs=task2_kwargs),
    ])
    assert [(0, result1), (1, result2)] == obj.receive_all()
    assert [2, 3] == obj.put_multiple([
        dict(task=task3, args=task3_args, kwargs=task3_kwargs),
        dict(task=task4, args=task4_args, kwargs=task4_kwargs),
    ])
    assert [(2, result3), (3, result4)] == obj.receive_all()

def test_put_multiple_receive_all_end_repeat(obj):
    assert [0, 1] == obj.put_multiple([
        dict(task=task1, args=task1_args, kwargs=task1_kwargs),
        dict(task=task2, args=task2_args, kwargs=task2_kwargs),
    ])
    assert [(0, result1), (1, result2)] == obj.receive_all()
    obj.end()
    obj.begin()
    assert [2, 3] == obj.put_multiple([
        dict(task=task3, args=task3_args, kwargs=task3_kwargs),
        dict(task=task4, args=task4_args, kwargs=task4_kwargs),
    ])
    assert [(2, result3), (3, result4)] == obj.receive_all()

##__________________________________________________________________||
def test_receive_finished_four_tasks(obj):
    assert [0, 1, 2, 3] == obj.put_multiple([
        dict(task=task1, args=task1_args, kwargs=task1_kwargs),
        dict(task=task2, args=task2_args, kwargs=task2_kwargs),
        dict(task=task3, args=task3_args, kwargs=task3_kwargs),
        dict(task=task4, args=task4_args, kwargs=task4_kwargs),
    ])

    runid_result_pairs = [ ]
    ntasks_to_wait = 4
    while len(runid_result_pairs) < ntasks_to_wait:
        runid_result_pairs.extend(obj.receive_finished())

    assert sorted([(0, result1), (1, result2), (2, result3), (3, result4)]) == sorted(runid_result_pairs)

##__________________________________________________________________||
def test_receive_finished_then_receive_all(obj):
    assert [0, 1, 2, 3] == obj.put_multiple([
        dict(task=task1, args=task1_args, kwargs=task1_kwargs),
        dict(task=task2, args=task2_args, kwargs=task2_kwargs),
        dict(task=task3, args=task3_args, kwargs=task3_kwargs),
        dict(task=task4, args=task4_args, kwargs=task4_kwargs),
    ])

    runid_result_pairs = [ ]

    # receive the first two with receive_finished()
    ntasks_to_wait = 2
    while len(runid_result_pairs) < ntasks_to_wait:
        runid_result_pairs.extend(obj.receive_finished())

    # the rest with receive_all()
    runid_result_pairs.extend(obj.receive_all())

    assert sorted([(0, result1), (1, result2), (2, result3), (3, result4)]) == sorted(runid_result_pairs)

##__________________________________________________________________||
def test_receive_one_four_tasks(obj):
    assert [0, 1, 2, 3] == obj.put_multiple([
        dict(task=task1, args=task1_args, kwargs=task1_kwargs),
        dict(task=task2, args=task2_args, kwargs=task2_kwargs),
        dict(task=task3, args=task3_args, kwargs=task3_kwargs),
        dict(task=task4, args=task4_args, kwargs=task4_kwargs),
    ])

    runid_result_pairs = [ ]
    runid_result_pairs.append(obj.receive_one())
    runid_result_pairs.append(obj.receive_one())
    runid_result_pairs.append(obj.receive_one())
    runid_result_pairs.append(obj.receive_one())

    assert obj.receive_one() is None

    assert sorted([(0, result1), (1, result2), (2, result3), (3, result4)]) == sorted(runid_result_pairs)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import time
from operator import itemgetter
from collections import deque

from .WorkingArea import WorkingArea

##__________________________________________________________________||
class TaskPackageDropbox(object):
    """A drop box for task packages.

    It puts task packages in a working area and dispatches runners
    that execute the tasks.

    """
    def __init__(self, workingArea, dispatcher, sleep=5):
        self.workingArea = workingArea
        self.dispatcher = dispatcher
        self.sleep = sleep

    def __repr__(self):
        name_value_pairs = (
            ('workingArea', self.workingArea),
            ('dispatcher', self.dispatcher),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def open(self):
        self.workingArea.open()
        self.runid_pkgidx_map = { }
        self.runid_to_return = deque() # finished runids

    def put(self, package):
        """Put a package. Return a package index.
        """

        pkgidx = self.workingArea.put_package(package)

        logger = logging.getLogger(__name__)
        logger.info('submitting {}'.format(self.workingArea.package_path(pkgidx)))

        runid = self.dispatcher.run(self.workingArea, pkgidx)
        self.runid_pkgidx_map[runid] = pkgidx
        return pkgidx

    def put_multiple(self, packages):
        """Put multiple packages. Return package indices.
        """
        pkgidxs = [self.workingArea.put_package(p) for p in packages]

        logger = logging.getLogger(__name__)
        logger.info('submitting {}'.format(
            ', '.join(['{}'.format(self.workingArea.package_path(i)) for i in pkgidxs])
        ))

        runids = self.dispatcher.run_multiple(self.workingArea, pkgidxs)
        self.runid_pkgidx_map.update(zip(runids, pkgidxs))
        return pkgidxs

    def poll(self):
        """Return pairs of package indices and results of finished tasks.
        """
        pkgidx_result_pairs = self._collect_all_finished_pkgidx_result_pairs()

        # remove failed results and sort in the order of pkgidx
        pkgidx_result_pairs = filter(itemgetter(1), pkgidx_result_pairs)
        pkgidx_result_pairs = sorted(pkgidx_result_pairs, key=itemgetter(0))

        return pkgidx_result_pairs

    def receive_one(self):
        """Return a pair of a package index and result.

        This method waits until a task finishes.
        This method returns None if no task is running.
        """
        if not self.runid_pkgidx_map:
            return None

        pkgidx_result_pair = None
        while not pkgidx_result_pair:

            pkgidx_result_pair = self._collect_next_finished_pkgidx_result_pair()

            if pkgidx_result_pair and not pkgidx_result_pair[1]:
                pkgidx_result_pair = None

            # Only sleep if there are no pending runs
            if not self.runid_to_return and self.runid_pkgidx_map:
                time.sleep(self.sleep)

        return pkgidx_result_pair

    def receive(self):
        """Return pairs of package indices and results.

        This method waits until all tasks finish.
        """

        pkgidx_result_pairs = []
        while self.runid_pkgidx_map:
            pkgidx_result_pairs.extend(
                self._collect_all_finished_pkgidx_result_pairs()
            )

            # Only sleep if we're waiting for runs
            if self.runid_pkgidx_map:
                time.sleep(self.sleep)

        # remove failed results and sort in the order of pkgidx
        pkgidx_result_pairs = filter(itemgetter(1), pkgidx_result_pairs)
        pkgidx_result_pairs = sorted(pkgidx_result_pairs, key=itemgetter(0))

        return pkgidx_result_pairs

    def _collect_all_finished_pkgidx_result_pairs(self):
        pkgidx_result_pairs = []

        pairs = self._collect_next_finished_pkgidx_result_pair()
        while self.runid_to_return:
            if pairs: pkgidx_result_pairs.append(pairs)
            pairs = self._collect_next_finished_pkgidx_result_pair()
        if pairs: pkgidx_result_pairs.append(pairs)

        return pkgidx_result_pairs

    def _collect_next_finished_pkgidx_result_pair(self):
        if not self.runid_to_return:
            self.runid_to_return.extend(self.dispatcher.poll())

        # No finished runs remaining
        if not self.runid_to_return:
            return None

        runid = self.runid_to_return.popleft()
        pkgidx = self.runid_pkgidx_map.pop(runid)
        result = self.workingArea.collect_result(pkgidx)

        # rerun failed job
        if result is None:
            logger = logging.getLogger(__name__)
            logger.warning('resubmitting {}'.format(
                self.workingArea.package_path(pkgidx)
            ))
            self.dispatcher.failed_runids([runid])
            runid = self.dispatcher.run(self.workingArea, pkgidx)
            self.runid_pkgidx_map[runid] = pkgidx

        return (pkgidx, result)

    def terminate(self):
        self.dispatcher.terminate()

    def close(self):
        self.workingArea.close()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import os

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.misc import profile_func, print_profile_func

##__________________________________________________________________||
def test_profile_func():
    func = mock.Mock()
    profile_func(func)

##__________________________________________________________________||
def test_print_profile_func(tmpdir_factory):
    tmpdir = str(tmpdir_factory.mktemp(''))
    func = mock.Mock()
    print(tmpdir)
    profile_out_path = os.path.join(tmpdir, 'profile.txt')
    print_profile_func(func, profile_out_path=profile_out_path)

##__________________________________________________________________||
